import copy

import gwpy
import numpy as np
from scipy.interpolate import interp1d
from scipy.signal import get_window

from pygwb.constants import H0


def parse_window_dict(window_dict):
    '''
    Pasrse the window dictionary properly for scipy compatibility.
    '''
    bools = ['sym', 'norm']
    floats = ['center', 'tau', 'alpha', 'beta', 'nbar', 'sll', 'std', 'p', 'sig', 'at']
    for key in window_dict.keys():
        if key in floats:
            window_dict[key] = float(window_dict[key])
        elif key in bools:
            window_dict[key] = bool(window_dict[key])
        else:
            pass
    return window_dict

def window_factors(N, window_fftgram_dict={"window_fftgram": "hann"}, overlap_factor=0.5):
    """
    Calculate window factors. By default, for a hann window with 50% overlap.

    Parameters:
    ===========
    window_fftgram_dict: dictionary, optional
        Dictionary with window characteristics. Default is `(window_fftgram_dict={"window_fftgram": "hann"}`
    overlap_factor: float, optional
        Defines the overlap between consecutive data chunks used in the calculation. Default is 0.5.
        

    Returns:
    ========
    w1w2bar: float
    w1w2squaredbar: float
    w1w2ovlbar: float
    w1w2squaredovlbar: float
    """
    window_tuple = get_window_tuple(window_fftgram_dict)
    w = get_window(window_tuple, N, fftbins=False)
    w1w2bar = np.mean(w ** 2)
    w1w2squaredbar = np.mean(w ** 4)

    S = N - int(overlap_factor*N)
    
    if overlap_factor==0.0 or overlap_factor==0:
        w1w2squaredovlbar = 0.0
        w1w2ovlbar = 0.0
    else:
        w1w2squaredovlbar = 1 / (N*overlap_factor) * np.sum(w[0:N-S]**2*w[S:N]**2)
        w1w2ovlbar = 1 / (N*overlap_factor) * np.sum(w[0:N-S]*w[S:N])

    return w1w2bar, w1w2squaredbar, w1w2ovlbar, w1w2squaredovlbar


def get_window_tuple(window_fftgram_dict={"window_fftgram": "hann"}):
    """
    Unpack the `window_fft_dict` dictionary into a `tuple` that may be read by scipy.get_window.

    Parameters:
    ===========
    window_fftgram_dict: dictionary, optional
        Dictionary with window characteristics. Default is `(window_fftgram_dict={"window_fftgram": "hann"}`.

    Returns:
    ========
    window_tuple: tuple
        A tuple containing the window_fft name as the first entry, followed by optional entries of the window_fft_dict.

    Notes:
    ======
    `window_fftgram_dict` is expected to have at least one item, `window_fftgram`.
    """
    window_dict = copy.deepcopy(window_fftgram_dict)
    out = tuple([window_dict["window_fftgram"]])
    window_dict.pop("window_fftgram")
    for name in window_dict:
        if name != "sym":
            out += tuple([window_dict[name]])
    if "sym" in window_fftgram_dict:
        out += tuple([window_dict["sym"]])
    return out


def calc_rho1(N, window_fftgram_dict={"window_fftgram": "hann"}, overlap_factor=0.5):
    """
    Calculate the combined window factor rho.

    Parameters:
    ===========
    N: int
        Length of the window.
    window_fftgram_dict: dictionary, optional
        Dictionary with window characteristics. Default is `(window_fftgram_dict={"window_fftgram": "hann"}`.
    overlap_factor: float, optional
        Defines the overlap between consecutive data chunks used in the calculation. Default is 0.5.

    Returns:
    ========
    rho1: float
        The combined window factor.
    """
    w1w2bar, _, w1w2ovlbar, _ = window_factors(N, window_fftgram_dict, overlap_factor=overlap_factor)
    rho1 = (overlap_factor * w1w2ovlbar / w1w2bar) ** 2
    return rho1


def calc_rho(N, j, window_tuple="hann", overlap_factor=0.5):
    """
    Calculate the normalised correlation of a window with itself shifted j times. This is identical
    to the rho(j) from Welch (1967).

    Parameters:
    ===========
    N: int
        Length of the window.
    j: int
        Number of "shifts" to apply to the window when correlating with itself.
    window_tuple: string or tuple, optional
        Window name or tuple as used in get_window(). Default is `window_tuple="hann"`.
    overlap_factor: float, optional
        Defines the overlap between consecutive segments used in the calculation. Default is 0.5.

    Returns:
    ========
    rho: float
        The normalised window correlation rho(j).
    """
    # The base window for which we want to calculate the correlation
    w = get_window(window_tuple, N, fftbins=False)
    
    # S is the shift, that is, the number of samples by which the window is shifted from the base window
    S = N - int(overlap_factor*N)
    
    if (j*S < N):
        rho = (np.sum(w[0:N-j*S]*w[j*S:N])/sum(w*w))**2
    else:
        # j*S >= N so no overlap
        rho = 0

    return rho


def effective_welch_averages(nSamples, N, window_tuple="hann", overlap_factor=0.5):
    """
    Calculate the "effective" number of averages used in Welch's PSD estimate after taking into account windowing
    and overlap.

    Parameters:
    ===========
    nSamples: int
        Number of samples to be used to estimate the PSD.
    N: int
        Length of the window.
    window_tuple: string or tuple, optional
        Window name or tuple as used in get_window(). Default is `window_tuple="hann"`.
    overlap_factor: float, optional
        Defines the overlap between consecutive segments used in the calculation. Default is 0.5.

    Returns:
    ========
    Neff: float
        The effective number of averages.
    """
    # S is the shift, that is, the number of samples by which the window is shifted from the base window
    S = N - int(overlap_factor*N)
    
    # K is the number of segments that will be averaged in the corresponding Welch estimate
    K = 1 + int((nSamples - N)/S)

    # Form the weighted sum of the window correlations for shifts j = 1 to K-1
    rho_sum = 0
    for j in range(1, K):
        rho_sum += (K - j)/K*calc_rho(N, j, window_tuple=window_tuple, overlap_factor=overlap_factor)
    
    Neff = K/(1 + 2 * rho_sum)

    return Neff


def calc_bias(
    segmentDuration,
    deltaF,
    deltaT,
    N_avg_segs=2,
    window_fftgram_dict={"window_fftgram": "hann"},
    overlap_factor=0.5
):
    """
    Calculate the bias factor introduced by welch averaging.

    Parameters:
    ===========
    segmentDuration: float
        Duration in seconds of welched segment.
    deltaF: float
        Frequency resolution of welched segment.
    deltaT: float
        Time sampling of welched segment.
    N_avg_segs: int, optional
        Number of segments over which the average is performed.
    overlap_factor: float, optional
        Defines the overlap between consecutive data chunks used in the calculation. Default is 0.5.

    Returns:
    ========
    bias: float
        The bias factor.
    """
    # Number of samples in a data chunk
    nSamples = int(segmentDuration / deltaT)

    # Number of samples in the window used for Welch's estimate
    N = int(1 / (deltaT * deltaF))

    # Effective number of segments that are averaged for this windowing scheme
    window_tuple = get_window_tuple(window_fftgram_dict)
    Neff = effective_welch_averages(nSamples, N, window_tuple, overlap_factor=overlap_factor)

    # if Neff is the same as the true number of segments, no bias.
    if Neff == (nSamples/N):
        bias = 1.0

    else:
        # Correction for number of PSDs that will be averaged 
        Neff = N_avg_segs * Neff
        bias = Neff / (Neff - 1)

    return bias


def omega_to_power(omega_GWB, frequencies):
    """
    Compute the GW power spectrum starting from the omega_GWB
    spectrum.

    Parameters:
    ===========
    omega_GWB: array_like
        The omega spectrum to turn into strain power.
    frequencies: array_like
        Array of frequencies corresponding to the omega spectrum.

    Returns:
    ========
    power: gwpy.frequencyseries.FrequencySeries
        A gwpy FrequencySeries containing the GW power spectrum

    Notes:
    ======
    The given frequencies need to match the given spectrum.
    """
    H_theor = (3 * H0.si.value ** 2) / (10 * np.pi ** 2)

    power = H_theor * omega_GWB * frequencies ** (-3)
    power = gwpy.frequencyseries.FrequencySeries(power, frequencies=frequencies)

    return power


def interpolate_frequency_series(fSeries, new_frequencies):
    """
    Interpolate a frequency series, given a new set of frequencies.

    Parameters:
    ===========
    fSeries: gwpy.frequencyseries.FrequencySeries
        The fFrequencySeries to interpolate.
    new_frequencies: array_like
        The new set of frequencies to interpolate to.

    Returns:
    ========
    fSeries_new: gwpy.frequencyseries.FrequencySeries
        The interpolated FrequencySeries.

    """
    spectrum = fSeries.value
    frequencies = fSeries.frequencies.value

    spectrum_func = interp1d(
        frequencies, spectrum, kind="cubic", fill_value="extrapolate"
    )

    return gwpy.frequencyseries.FrequencySeries(
        spectrum_func(new_frequencies), frequencies=new_frequencies
    )


def StatKS(DKS):
    """
    Compute the KS test.
    """
    jmax = 500
    pvalue = 0.0
    for jj in np.arange(1, jmax + 1):
        pvalue += 2.0 * (-1) ** (jj + 1) * np.exp(-2.0 * jj ** 2 * DKS ** 2)
    return pvalue

def _check_omegaspectra(spectra):
        for spec in spectra:
            if spec.alpha != spectra[0].alpha:
                print(spec.alpha, spectra[0].alpha)
                raise ValueError('spectra in this set have been weighted with different alphas. Please correct this before continuing.')
            if spec.fref != spectra[0].fref:
                raise ValueError('spectra in this set have been set at different reference frequencies. Please correct this before continuing.')
            if spec.h0 != spectra[0].h0:
                raise ValueError('spectra in this set have been set at different h0. Please correct this before continuing.')
            if not np.allclose(spec.frequencies.value, spectra[0].frequencies.value):
                raise ValueError('spectra in this set have different frequencies. Please correct this before continuing.')



import numpy as np
from bilby.gw.detector.strain_data import Notch


class StochNotch(Notch):
    def __init__(self, minimum_frequency, maximum_frequency, description):
        """A notch object storing the maximum and minimum frequency of the notch, as well as a description

        Parameters
        ==========
        minimum_frequency, maximum_frequency: float
            The minimum and maximum frequency of the notch
        description: str
            A description of the origin/reason of the notch

        """
        super().__init__(minimum_frequency, maximum_frequency)
        self.description = description

    def print_notch(self):
        print(self.minimum_frequency, self.maximum_frequency, self.description)

    def get_notch_mask(self, frequency_array):
        """Get a boolean mask for the frequencies in frequency_array in the notch

        Parameters
        ==========
        frequency_array: np.ndarray
            An array of frequencies

        Returns
        =======
        notch_mask: np.ndarray
            An array of booleans that are False for frequencies in the notch

        Notes
        =====
        This notches any frequency that may have overlapping frequency content with the notch.
        """
        df = np.abs(frequency_array[1] - frequency_array[0])
        frequencies_below = np.concatenate(
            [frequency_array[:1] - df, frequency_array[:-1]]
        )
        frequencies_above = np.concatenate(
            [frequency_array[1:], frequency_array[-1:] + df]
        )
        lower = frequencies_below + df / 2 <= self.maximum_frequency
        upper = frequencies_above - df / 2 >= self.minimum_frequency
        notch_mask = [not elem for elem in (lower & upper)]
        return notch_mask


class StochNotchList(list):
    def __init__(self, notch_list):
        """A list of notches

        Parameters
        ==========
        notch_list: list
            A list of length-3 tuples of the (min, max) frequency; description for the notches.

        Raises
        ======
        ValueError
            If the list is malformed.
        """

        if notch_list is not None:
            for notch in notch_list:
                if isinstance(notch, tuple) and len(notch) == 3:
                    self.append(StochNotch(*notch))
                else:
                    msg = f"notch_list {notch_list} is malformed"
                    raise ValueError(msg)

    def check_frequency(self, freq):
        """Check if freq is inside the notch list

        Parameters
        ==========
        freq: float
            The frequency to check

        Returns
        =======
        True/False:
            If freq inside any of the notches, return True, else False
        """

        for notch in self:
            if notch.check_frequency(freq):
                return True
        return False

    def get_notch_mask(self, frequency_array, save_file_flag=False, filename=""):
        """Get a boolean mask for the frequencies in frequency_array in the notch list

        Parameters
        ==========
        frequency_array: np.ndarray
            An array of frequencies
        save_file_flag: boolean
            A boolean flag indicating whether to save the notch mask in a file or not
        filename: str
            The name of the file where to store the notch mask if save_file_flag is true

        Returns
        =======
        notch_mask: np.ndarray
            An array of booleans that are False for frequencies in the notch

        Notes
        =====
        This notches any frequency that may have overlapping frequency content with the notch.
        """
        notch_mask = np.ones(len(frequency_array), dtype=bool)
        for notch in self:
            notch_mask = notch_mask & notch.get_notch_mask(frequency_array)

        if save_file_flag == True:
            if len(filename) == 0:
                filename = "Notch_mask.txt"
            save_notch_mask(self, frequency_array, filename)
        return notch_mask

    def save_notch_mask(self, frequency_array, filename):
        """Saves a boolean mask for the frequencies in frequency_array in the notch list

        Parameters
        ==========
        frequency_array: np.ndarray
            An array of frequencies

        filename: str
            Name of the target file

        Notes
        =====
        This saves notch_mask (see get_notch_mask for more information) in a text file. notch_mask is an array of booleans that are False for frequencies in the notch.
        """

        notch_mask = self.get_notch_mask(frequency_array)
        np.savetxt(
            filename,
            np.transpose([frequency_array, notch_mask]),
        )

    def save_to_txt(self, filename):
        """Save the nocth list to a txt-file (after sorting)

        Parameters
        ==========
        filename: str
            Name of the target file

        """

        fmin = []
        fmax = []
        desc = []
        self.sort_list()
        for n in self:
            fmin.append(n.minimum_frequency)
            fmax.append(n.maximum_frequency)
            desc.append(n.description)

        np.savetxt(
            filename,
            np.transpose([fmin, fmax, desc]),
            fmt=("%-20s  ,  %-20s  ,  %-" + str(len(max(desc)) + 5) + "s"),
        )

    def sort_list(self):
        """Sorts the notch list based on the minimum frequency of the notches

        Parameters
        ==========

        """

        self.sort(key=lambda elem: elem.minimum_frequency)

    @classmethod
    def load_from_file(cls, filename):
        """Load an already existing notch list from a txt-file (with formatting as produced by this code)

        Parameters
        ==========
        filename: str
            Filename of the file containing the notchlist to be read in

        """

        fmin, fmax = np.loadtxt(filename, delimiter=",", unpack=True, usecols=(0, 1))
        desc = np.loadtxt(
            filename, delimiter=",", unpack=True, usecols=(2), dtype="str"
        )

        cls = StochNotchList([])
        if np.ndim(fmin) == 1:
            for i in range(len(fmin)):
                cls.append(StochNotch(fmin[i], fmax[i], desc[i]))
        elif np.ndim(fmin) == 0:
            cls.append(StochNotch(fmin, fmax, desc))
        else:
            raise TypeError("Notch list from file has too many dimensions.")
        return cls


def power_lines(fundamental=60, nharmonics=40, df=0.2):
    """
    Create list of power line harmonics (nharmonics*fundamental Hz) to remove

    Parameters
    ==========
    fundamental: float
        Fundamental frequency of the first harmonic
    nharmonics: float
        Number of harmonics (should include all harmonics within studied frequency range of the study)

    Returns
    =======
    notches: list of NoiseLine objects
        List of lines you want to be notched in NoisLine format

    """
    freqs = fundamental * np.arange(1, nharmonics + 1)

    notches = StochNotchList([])
    for f0 in freqs:
        notch = StochNotch(f0 - df / 2, f0 + df / 2, "Power Lines")
        notches.append(notch)

    return notches


def comb(f0, f_spacing, n_harmonics, df, description=None):
    """
    Create a list of comb lines to remove with the form 'f0+n*f_spacing, n=0,1,...,n_harmonics-1'

    Parameters
    ==========
    f0: float
        Fundamental frequency of the first harmonic
    f_spacing: float
        spacing between two subsequent harmonics
    nharmonics: float
        Number of harmonics (should include all harmonics within studied frequency range of the study)
    df: float
        Width of the comb-lines
    description: str (Optional)
        Optional additional description, e.g. known source of the comb

    Returns
    =======
    notches: list of NoiseLine objects
        List of lines you want to be notched in NoisLine format

    """

    notches = StochNotchList([])
    freqs = [f0 + n * f_spacing for n in range(n_harmonics)]
    for f in freqs:
        TotalDescription = f"Comb with fundamental freq {f0} and spacing {f_spacing}"
        if description:
            TotalDescription += " " + description
        notch = StochNotch(f - df / 2, f + df / 2, TotalDescription)
        notches.append(notch)

    return notches


def pulsar_injections(filename, t_start, t_end, doppler=1e-4):
    """
    Create list of frequencies contaminated by pulsar injections

    Parameters
    ==========
    filename: str
        Filename of list containing information about pulsar injections. e.g. for O3 at https://git.ligo.org/stochastic/stochasticdetchar/-/blob/master/O3/notchlists/make_notchlist/input/pulsars.dat
    t_start: int
        GPS start time of run/analysis
    t_end: int
        GPS end time of run/analysis
    doppler: float
        Doppler shift; typical value of v/c for Earth motion in solar system = 1e-4 (default)

    Returns
    =======
    notches: list of NoiseLine objects
        List of lines you want to be notched in NoisLine format
    """

    """
    f_start: pulsar freq at start of time period
    f_end:   pulsar freq at end of time period
    f1:      allow for doppler shifting
    f2:      allow for doppler shifting
    f0:      central freq over entire period
    df:      width
    binary:  pulsar binary system, yes or no. If yes the affected with is ~two times larger (by design). We use a conservative factor of 3. 
    """

    t_refs, f_refs, f_dots, binary = np.loadtxt(
        filename,
        unpack=True,
        dtype=[
            ("t_refs", float),
            ("f_refs", float),
            ("f_dots", float),
            ("binary", str, 3),
        ],
    )
    notches = StochNotchList([])

    for t_ref, f_ref, f_dot, my_binary in zip(t_refs, f_refs, f_dots, binary):
        f_start = f_ref + f_dot * (t_start - t_ref)
        f_end = f_ref + f_dot * (t_end - t_ref)
        f1 = f_start * (1 + doppler)
        f2 = f_end * (1 - doppler)
        f0 = (f1 + f2) / 2.0
        df = f1 - f2
        if my_binary == "yes":
            df = 3.0 * df
        notch = StochNotch(f0 - df / 2, f0 + df / 2, "Pulsar injection")
        notches.append(notch)
    return notches



import numpy as np


def calculate_coherence(psd_1, psd_2, csd):
    '''
    Calculate a coherence spectrum given two power spectrograms and a cross-spectrogram.

    Parameters
    ==========
    psd_1_spectrum: np.array
        PSD spectrum of first detector.
    psd_2_spectrum: np.array
        PSD spectrum of second detector.
    csd_spectrum: np.array
        Spectrum of the cross-spectral density of the two detectors.
    '''
    #I don't think we need these terms anymore:
    #fftlength = int(1.0/psd_1_spectrogram.df.value)
    #norm1 = int(np.floor(duration/(fftlength*overlap_factor))-1)
    return  np.real(csd*np.conj(csd)/(psd_1*psd_2))




import numpy as np
from scipy.special import spherical_jn

from .constants import speed_of_light


def Tplus(alpha, beta):
    return (
        -(
            3.0 / 8 * spherical_jn(0, alpha)
            - 45.0 / 56 * spherical_jn(2, alpha)
            + 169.0 / 896 * spherical_jn(4, alpha)
        )
        + (
            0.5 * spherical_jn(0, alpha)
            - 5.0 / 7 * spherical_jn(2, alpha)
            - 27.0 / 224 * spherical_jn(4, alpha)
        )
        * np.cos(beta)
        - (
            1.0 / 8 * spherical_jn(0, alpha)
            + 5.0 / 56 * spherical_jn(2, alpha)
            + 3.0 / 896 * spherical_jn(4, alpha)
        )
        * np.cos(2 * beta)
    )


def Tminus(alpha, beta):
    return (
        spherical_jn(0, alpha)
        + 5.0 / 7 * spherical_jn(2, alpha)
        + 3.0 / 112 * spherical_jn(4, alpha)
    ) * np.cos(beta / 2) ** 4


def Vplus(alpha, beta):
    return (
        -(
            3.0 / 8 * spherical_jn(0, alpha)
            + 45.0 / 112 * spherical_jn(2, alpha)
            + 169.0 / 224 * spherical_jn(4, alpha)
        )
        + (
            0.5 * spherical_jn(0, alpha)
            + 5.0 / 14 * spherical_jn(2, alpha)
            + 27.0 / 56 * spherical_jn(4, alpha)
        )
        * np.cos(beta)
        - (
            1.0 / 8 * spherical_jn(0, alpha)
            - 5.0 / 112 * spherical_jn(2, alpha)
            - 3.0 / 224 * spherical_jn(4, alpha)
        )
        * np.cos(2 * beta)
    )


def Vminus(alpha, beta):
    return (
        spherical_jn(0, alpha)
        - 5.0 / 14 * spherical_jn(2, alpha)
        - 3.0 / 28 * spherical_jn(4, alpha)
    ) * np.cos(beta / 2) ** 4


def Splus(alpha, beta):
    return (
        -(
            3.0 / 8 * spherical_jn(0, alpha)
            + 45.0 / 56 * spherical_jn(2, alpha)
            + 507.0 / 448 * spherical_jn(4, alpha)
        )
        + (
            0.5 * spherical_jn(0, alpha)
            + 5.0 / 7 * spherical_jn(2, alpha)
            - 81.0 / 112 * spherical_jn(4, alpha)
        )
        * np.cos(beta)
        - (
            1.0 / 8 * spherical_jn(0, alpha)
            - 5.0 / 56 * spherical_jn(2, alpha)
            + 9.0 / 448 * spherical_jn(4, alpha)
        )
        * np.cos(2 * beta)
    )


def Sminus(alpha, beta):
    return (
        spherical_jn(0, alpha)
        - 5.0 / 7 * spherical_jn(2, alpha)
        + 9.0 / 56 * spherical_jn(4, alpha)
    ) * np.cos(beta / 2) ** 4


def T_right_left(alpha, beta):
    return -np.sin(beta / 2) * (
        (-spherical_jn(1, alpha) + 7.0 / 8 * spherical_jn(3, alpha))
        + (spherical_jn(1, alpha) + 3.0 / 8 * spherical_jn(3, alpha)) * np.cos(beta)
    )


def tangent_vector(vector1, vector2):
    return np.subtract(
        vector2,
        np.multiply(np.dot(vector1, vector2) / np.dot(vector1, vector1), vector1),
    )


def omega_tangent_bisector(bisector, tangent_vector, perp):
    norm = np.linalg.norm(bisector) * np.linalg.norm(tangent_vector)
    sin_omega = np.dot(np.cross(bisector, tangent_vector), perp) / norm
    cos_omega = np.dot(bisector, tangent_vector) / norm
    return np.arctan2(sin_omega, cos_omega)


def calc_orf(
    frequencies,
    det1_vertex,
    det2_vertex,
    det1_xarm,
    det2_xarm,
    det1_yarm,
    det2_yarm,
    polarization="tensor",
):
    """
    Calculates the tensor, scalar, and vector overlap reduction functions
    Following Section IVb of https://arxiv.org/abs/0903.0528
    See Appendix A of https://arxiv.org/abs/1704.08373 for a
    discussion of the normalization of the scalar ORF and
    https://arxiv.org/pdf/0707.0535.pdf for the gamma_V function

    Inputs:
    frequencies: frequencies at which to evaluate the ORFs
    det1_vertex: Coordinates of the vertex of detector 1
    det2_vertex: Coordinates of the vertex of detector 2
    det1_xarm: Coordinates of the x arm of detector 1
    det2_xarm: Coordinates of the x arm of detector 2
    det1_yarm: Coordinates of the y arm of detector 1
    det2_yarm: Coordinates of the y arm of detector 2
    Coordinates are always Earth-fixed cartesian

    Description of the intermediate parameters:
    beta: angle between detectors from center of earth
    tan_detX: tangent vector at detX along great circle between detectors
    bisector_detX: detX arm bisector vector
    perp_detX: outward radial vector perpendicular to the detector plane
    omega_detX: angle between bisector and tangent vector at detX
    perp: vector at theta=90 along great circle with det1_vertex theta=0

    Outputs:
    overlap_reduction_function: overlap reduction function at given frequencies
        for specified polarization
    """

    delta_x = np.subtract(det1_vertex, det2_vertex)
    alpha = 2 * np.pi * frequencies * np.linalg.norm(delta_x) / speed_of_light

    beta = np.arccos(
        np.dot(det1_vertex, det2_vertex)
        / (np.linalg.norm(det1_vertex) * np.linalg.norm(det2_vertex))
    )

    tan_det1 = tangent_vector(det1_vertex, det2_vertex)
    bisector_det1 = np.add(det1_xarm, det1_yarm)
    perp_det1 = -np.cross(det1_xarm, det1_yarm) / (
        np.linalg.norm(det1_xarm) * np.linalg.norm(det1_yarm)
    )

    perp = np.cross(np.cross(det1_vertex, det2_vertex), det1_vertex)
    tan_det2 = tangent_vector(det2_vertex, perp)
    bisector_det2 = np.add(det2_xarm, det2_yarm)
    perp_det2 = -np.cross(det2_xarm, det2_yarm) / (
        np.linalg.norm(det2_xarm) * np.linalg.norm(det2_yarm)
    )

    if np.linalg.norm(delta_x) != 0:
        omega_det1 = omega_tangent_bisector(bisector_det1, tan_det1, perp_det1)
        omega_det2 = omega_tangent_bisector(bisector_det2, tan_det2, perp_det2)
        omega_plus = (omega_det1 + omega_det2) / 2
        omega_minus = (omega_det1 - omega_det2) / 2
    else:
        omega_plus = 1
        omega_minus = (
            omega_tangent_bisector(bisector_det1, bisector_det2, perp_det1) / 2
        )

    if polarization.lower() == "tensor":
        overlap_reduction_function = Tplus(alpha, beta) * np.cos(
            4 * omega_plus
        ) + Tminus(alpha, beta) * np.cos(4 * omega_minus)
    elif polarization.lower() == "vector":
        overlap_reduction_function = Vplus(alpha, beta) * np.cos(
            4 * omega_plus
        ) + Vminus(alpha, beta) * np.cos(4 * omega_minus)
    elif polarization.lower() == "scalar":
        overlap_reduction_function = (
            1.0
            / 3
            * (
                Splus(alpha, beta) * np.cos(4 * omega_plus)
                + Sminus(alpha, beta) * np.cos(4 * omega_minus)
            )
        )
    elif polarization.lower() == "right_left":
        overlap_reduction_function = T_right_left(alpha, beta) * np.sin(4 * omega_plus)
    else:
        raise ValueError(
            "Unrecognized polarization! Must be either tensor, vector, scalar, or right_left"
        )
    return overlap_reduction_function



import copy
import warnings

import numpy as np
from gwpy.frequencyseries import FrequencySeries
from gwpy.spectrogram import Spectrogram
from gwpy.timeseries import TimeSeries
from scipy.signal import get_window, spectrogram

from pygwb.util import get_window_tuple, parse_window_dict


def fftgram(
    time_series_data: TimeSeries,
    fftlength: int,
    overlap_factor: float = 0.5,
    zeropad: bool = False,
    window_fftgram_dict: dict = {"window_fftgram": "hann"},
):
    """Create an fftgram from a timeseries

    Parameters
    ----------
    time_series_data: gwpy timeseries
        Timeseries from which to compute the fftgram.
    fftlength: int
        Length of each segment (in seconds) for computing FFT.
    overlap_factor: float, optional
        Factor of overlap between adjacent FFT segments (values range from 0
        to 1). Users should provide proper combination of overlap_factor and
        window_fftgram_dict. For \"hann\" window use 0.5 overlap_factor and
        for \"boxcar"\ window use 0 overlap_factor. Default 0.5 (50% overlap).
    zeropad: bool, optional
        Before doing FFT whether to zero pad the data equal to the length of
        FFT or not. Default is False.
    window_fftgram_dict: dictionary, optional
        Dictionary containing name and parameters describing which window to
        use for producing FFTs. Default is \"hann\".

    Returns
    -------
    data_fftgram: gwpy spectrogram (complex)
        fftgram containing several FFTs in a matrix format
    """

    sample_rate = int(1 / time_series_data.dt.value)

    # get the window function
    window_tuple = get_window_tuple(parse_window_dict(window_fftgram_dict))
    window_fftgram = get_window(window_tuple, fftlength * sample_rate, fftbins=False)

    # calculate the spectrogram using scipy.signal.spectrogram
    if zeropad:
        nfft = 2 * fftlength * sample_rate
    else:
        nfft = fftlength * sample_rate

    f, t, Sxx = spectrogram(
        time_series_data.data,
        fs=sample_rate,
        window=window_fftgram,
        nperseg=fftlength * sample_rate,
        noverlap=overlap_factor * fftlength * sample_rate,
        nfft=nfft,
        mode="complex",
        detrend=False,
    )

    # convert the above spectrogram into gwpy spectrogram object
    ffts = []
    for ii in range(len(t)):
        ffts.append(FrequencySeries(Sxx.T[ii], f0=f[0], df=f[1] - f[0]))

    data_fftgram = Spectrogram.from_spectra(
        *ffts, epoch=time_series_data.t0, dt=t[1] - t[0]
    )

    return data_fftgram


def pwelch_psd(
    psdgram: Spectrogram, segment_duration: int, overlap_factor: float = 0.5
):
    """
    Estimate PSD using pwelch method.

    Parameters
    ==========
    psdgram: gwpy spectrogram (PSD)
       PSD gram data to be averaged
    segment_duration: int
        Data duration over which PSDs need to be averaged. Should be greater
        than or equal to the duration used for FFT.
    overlap_factor: float, optional
        Amount of overlap between adjacent average PSDs, can vary between 0
        and 1. Default is 0.5.

    Returns
    =======
    avg_psdgram: gwpy psd spectrogram
        averaged over segments within the segment_duration
    """

    averaging_factor = round(segment_duration * psdgram.dy.value)

    if averaging_factor < 1:
        raise ValueError(
            """Segment_duration should be greater than the FFT
                        duration used for PSD calculation"""
        )
    elif averaging_factor == 1:  # nothing to average in this case
        avg_psdgram = np.real(psdgram)
    else:
        # total duration of the original time series
        job_duration = (
            psdgram.xindex.value[-1] + 1 / psdgram.dy.value - psdgram.xindex.value[0]
        )
        # possible number of (overlapping) segments
        stride = segment_duration * (1 - overlap_factor)
        n_segments = int((job_duration - overlap_factor * segment_duration) / stride)
        avg_psds = []
        segments_start_times = np.zeros(n_segments)
        start_time = psdgram.xindex.value[0]
        if n_segments==0:
            avg_psds = np.mean(psdgram[True], axis=0)
        for ii in range(n_segments):
            seg_indices = (psdgram.xindex.value >= start_time) & (
                psdgram.xindex.value
                <= (start_time + segment_duration - 1 / psdgram.dy.value)
            )
            avg_psds.append(psdgram[seg_indices].mean(axis=0))
            segments_start_times[ii] = start_time
            # move to next (overlapping) segment
            start_time = start_time + stride

        avg_psdgram = Spectrogram.from_spectra(
            *avg_psds, epoch=psdgram.xindex.value[0], dt=stride
        )

    return avg_psdgram


def before_after_average(psdgram: Spectrogram, segment_duration: int, N_avg_segs: int):
    """
    Average the requested number of PSDs from segments adjacent to the segment
    of interest (for which CDS is calculated)

    Parameters
    ----------
    psdgram: gwpy spectrogram (PSD)
        PSD spectrogram.
    segment_duration: int
        Duration of data used for each PSD calculation.
    N_avg_segs: int
        Number of segments to be used for PSD averaging (from both sides of
        the segment of interest). N_avg_segs should be even and >= 2

    Returns
    -------
    avg_psdgram: gwpy spectrogram
        averaged psd gram
    """

    if N_avg_segs < 2:
        raise ValueError("N_avg_segs should be >=2")

    if (N_avg_segs % 2) != 0:
        raise ValueError("N_avg_segs should be even")

    if (psdgram.xindex.value[-1] - psdgram.xindex.value[0]) < (
        N_avg_segs * segment_duration
    ):
        max_N_avg_segs = round(
            (psdgram.xindex.value[-1] - psdgram.xindex.value[0]) / segment_duration
        )
        raise ValueError(
            f""" Input (PSD) spectrogram does not have enough 
                           segments to be used with given N_avg_segs. Here 
                           N_avg_segs should be less than or equal to
                           {max_N_avg_segs}."""
        )

    # segments for which average PSDs need to be calculcated
    seg_times = [
        value
        for value in psdgram.xindex.value
        if (
            ((value - psdgram.xindex.value[0]) >= (N_avg_segs / 2 * segment_duration))
            & (
                (psdgram.xindex.value[-1] - value)
                >= (N_avg_segs / 2 * segment_duration)
            )
        )
    ]

    avg_psds = []
    for ii in seg_times:
        seg_diffs = abs(psdgram.xindex.value - ii) / segment_duration
        seg_indices = (
            (seg_diffs <= (N_avg_segs / 2))
            & ((seg_diffs % 1) == 0)
            & ((psdgram.xindex.value - ii) != 0)
        )
        # Below we only average non-overlapping segments (this is a choice)
        avg_psds.append(psdgram[seg_indices].mean(axis=0))

    avg_psdgram = Spectrogram.from_spectra(
        *avg_psds, epoch=seg_times[0], dt=psdgram.dx.value
    )
    return avg_psdgram


def coarse_grain(data, coarsening_factor):
    """
    Coarse grain a frequency series by an integer factor.

    If the coarsening factor is even, there are coarsening_factor + 1 entries
    in the input data that contribute to each coarse frequency bin, however,
    the first and last contribute only a half to the frequency below and half
    to the frequency above.

    If the coarsening factor is odd, there are no edge effects that have to be
    considered.

    The length of the output is len(data) // coarsening_factor - 1

    If the coarsening factor is not an integer, :code:`coarse_grain_exact` is
    used.

    Parameters
    ==========
    data: array-like
        The data to coarse grain
    coarsening_factor: float
        The factor by which to coarsen the data

    Returns
    =======
    coarsened: array-like
        The coarse-grained data
    """
    if coarsening_factor == 1:
        return data
    elif coarsening_factor % 1 != 0:
        return coarse_grain_exact(data, coarsening_factor)
    elif coarsening_factor % 2:
        data = data[:-1]
    coarsening_factor = int(coarsening_factor)
    coarsened = coarse_grain_naive(
        data=data[coarsening_factor // 2 + 1 : -(coarsening_factor // 2)],
        coarsening_factor=coarsening_factor,
    )
    if not coarsening_factor % 2:
        left_edges = data[coarsening_factor // 2 :: coarsening_factor][:-1]
        right_edges = data[int(coarsening_factor * 1.5) :: coarsening_factor]
        coarsened += (left_edges - right_edges) / 2 / coarsening_factor
    return coarsened


def coarse_grain_exact(data, coarsening_factor):
    """
    Coarse grain an array using any coarsening factor

    Each bin will contain the integral of the input array covering
    `coarsening_factor` bins.

    This is done by evaluating the difference between the cumulative integral
    of the data at the beginning and end of each bin.

    The i'th bin covers
    `[coarsening_factor * (ii - 0.5), coarsening_factor * (ii + 0.5)]`
    indexed for `1 <= ii < len(data) / coarsening_factor`.

    Parameters
    ==========
    data: array-like
        The data to coarse grain
    coarsening_factor: float
        The factor by which to coarsen the data

    Returns
    =======
    output: array-like
        The coarse-grained data
    """
    from scipy.integrate import cumtrapz

    n_input = len(data)
    first_full_bin_start = coarsening_factor / 2

    x_inputs = np.arange(n_input)
    x_values = np.arange(first_full_bin_start, n_input, coarsening_factor)

    cumulative_y = cumtrapz(data, x_inputs, initial=0)
    y_values = np.interp(x_values, x_inputs, cumulative_y)
    output = np.diff(y_values) / coarsening_factor

    return output


def coarse_grain_naive(data, coarsening_factor):
    """
    Naive implementation of a coarse graining factor that ignores edge effects.

    This is equivalent to the process performed for a Welch average

    Parameters
    ==========
    data: array-like
        The data to coarse grain
    coarsening_factor: int
        The factor by which to coarsen the data

    Returns
    =======
    coarsened:array-like
        The coarse-grained data
    """
    coarsening_factor = int(coarsening_factor)
    n_remove = len(data) % coarsening_factor
    if n_remove > 0:
        data = data[:-n_remove]
    coarsened = np.mean(data.reshape(-1, coarsening_factor), axis=-1)
    return coarsened


def coarse_grain_spectrogram(
    spectrogram,
    delta_t=None,
    delta_f=None,
    time_method="naive",
    frequency_method="full",
):
    """
    Coarsen a spectrogram in time and/or frequency, e.g., Welch averaging /
    coarse-graining

    The coarsening methods are either:
      - naive: this is equivalent to a Welch average
      - full: the full coarse-grain method
      - running_mean:

    Parameters
    ==========
    spectrogram: gwpy.spectrogram.Spectrogram
        Spectrogram object to be coarsened
    delta_t: float
        Output time spacing
    delta_f: float
        Output frequency spacing
    time_method: str
        Should be one of the coarsening methods listed above
    frequency_method: str
        Should be one of the coarsening methods listed above

    Returns
    =======
    output: gwpy.spectrogram.Spectrogram
    """
    from gwpy.spectrogram import Spectrogram

    methods = dict(
        naive=coarse_grain_naive,
        full=coarse_grain,
        running_mean=running_mean,
    )

    value = spectrogram.value

    if delta_t is not None:
        factor = delta_t / spectrogram.dt.value
        func = methods[time_method]
        value = np.apply_along_axis(func, axis=0, arr=value, coarsening_factor=factor)
        coarse_times = func(spectrogram.times.value, coarsening_factor=factor)
        coarse_times += spectrogram.times.value[0] - coarse_times[0]
    else:
        coarse_times = spectrogram.times

    if delta_f is not None:
        factor = delta_f / spectrogram.df.value
        func = methods[frequency_method]
        value = np.apply_along_axis(func, axis=1, arr=value, coarsening_factor=factor)
        coarse_frequencies = func(
            spectrogram.frequencies.value, coarsening_factor=factor
        )
    else:
        coarse_frequencies = spectrogram.frequencies

    # create gwpy spectrogram object
    specgrams = []
    # To avoid precision issues rounding the frequency values at ~1% level; ideally we expect the ratio coarse_frequencies[0]/spectrogram.dy.value/factor to be an integer
    f0 = np.round(coarse_frequencies[0]/spectrogram.dy.value/factor, 2) * spectrogram.dy.value * factor
    deltaF = f0
    for ii in value:
        specgrams.append(FrequencySeries(ii, f0=f0, df=deltaF))

    output = Spectrogram.from_spectra(
        *specgrams,
        epoch=coarse_times[0].value,
        dt=coarse_times[1].value - coarse_times[0].value,
    )

    return output


def cross_spectral_density(
    time_series_data1: TimeSeries,
    time_series_data2: TimeSeries,
    segment_duration: int,
    frequency_resolution: float,
    coarse_grain: bool = True,
    overlap_factor: float = 0.5,
    overlap_factor_welch: float = 0.5,
    zeropad: bool = False,
    window_fftgram_dict: dict = {"window_fftgram": "hann"},
    is_psd: bool = False,
):
    """
    Compute the cross spectral density from two time series inputs.

    Parameters
    ----------
    time_series_data1: gwpy timeseries
        Timeseries data of detector1.
    time_series_data2: gwpy timeseries
        Timeseries data of detector2.
    segment duration: int
        data duration over which CSDs need to be calculated.
    frequency_resolution: float
        Frequency resolution of the final CSDs. This is achieved by averaing in
        frequency domain.
    overlap_factor: float, optional
        Amount of overlap between adjacent segments (range between 0 and 1)
    overlap_factor_welch: float
        Overlap factor to use when using Welch's method (NOT coarsegraining). Users should provide proper combination of overlap_factor and window_fftgram_dict. For \"hann\" window use 0.5 overlap_factor and for \"boxcar"\ window use 0 overlap_factor. Default is 0.5 (50% overlap), which is optimal when using Welch's method with a \"hann\" window.
    zeropad: bool, optional
        Before doing FFT whether to zero pad the data equal to the length of
        FFT or not. Default is False.
    window_fftgram_dict: dictionary, optional
        Dictionary containing name and parameters describing which window to
        use for producing FFTs. Default is \"hann\".

    Returns
    -------
    csd_spectrogram: gwpy spectrogram
       Cross spectral density of the two timeseries
    """

    if (segment_duration * frequency_resolution) % 1 != 0:
        warnings.warn(
            "Multiplying the parameters segment_duration and frequency_resolution gives a non-integer. "
            "This could lead to possible issues while computing CSDs and PSDs. \n It will also lead to data loss when computing the PSD. "
            "To avoid this, make sure this multiplication is an integer."
        )

    if coarse_grain:
        fftlength = segment_duration
        overlap_factor_fftgram = overlap_factor
    else:
        # Length of data blocks to be used in pwelch
        fftlength = int(1.0 / frequency_resolution)
        overlap_factor_fftgram = overlap_factor_welch

    fft_gram_1 = fftgram(
        time_series_data1,
        fftlength,
        overlap_factor=overlap_factor_fftgram,
        zeropad=zeropad,
        window_fftgram_dict=window_fftgram_dict,
    )

    if is_psd:
        fft_gram_2 = copy.deepcopy(fft_gram_1)

    else:
        # Check if the lengths of two time-series are equal
        if len(time_series_data1.data) != len(time_series_data2.data):
            raise ValueError("Lengths of two input time series are not equal")

        # Check if the sample rates of two input time-series are equal
        if time_series_data1.dt.value != time_series_data2.dt.value:
            raise ValueError("Sample rates of two input time series are not equal")

        fft_gram_2 = fftgram(
            time_series_data2,
            fftlength,
            overlap_factor=overlap_factor_fftgram,
            zeropad=zeropad,
            window_fftgram_dict=window_fftgram_dict,
        )

    if not coarse_grain:
        csd_spectrogram = pwelch_psd(
            2 * np.conj(fft_gram_1) * fft_gram_2,
            segment_duration,
            overlap_factor=overlap_factor,
        )
    else:
        csd_spectrogram = coarse_grain_spectrogram(
            2 * np.conj(fft_gram_1) * fft_gram_2, delta_f=frequency_resolution
        )

    # Correct the scaling of DC and nyquist frequency components in agreement
    # with scipy.signal.welch (and pwelch in matlab)
    for ii in range(len(csd_spectrogram)):
        csd_spectrogram[ii].value[0] = csd_spectrogram[ii].value[0] / 2
        csd_spectrogram[ii].value[-1] = csd_spectrogram[ii].value[-1] / 2

    if is_psd:
        return np.real(csd_spectrogram)
    else:
        return csd_spectrogram


def power_spectral_density(
    time_series_data: TimeSeries,
    segment_duration: int,
    frequency_resolution: float,
    coarse_grain: bool = False,
    zeropad: bool = False,
    overlap_factor: float = 0.5,
    window_fftgram_dict: dict = {"window_fftgram": "hann"},
    overlap_factor_welch: float = 0.5,
):
    """
    Compute the PSDs of every segment (defined by the segment duration)
    in the time series using pwelch method.

    Parameters
    ----------
    time_series_data: gwpy timeseries
        Timeseries from which to compute PSDs.
    segment duration: int
        Data duration over which each PSDs need to be calculated.
    frequency_resolution: float
        Frequency resolution of the final PSDs. This sets the time duration
        over which FFTs are calculated in the pwelch method.
    coarse_grain: bool
        Coarse-graining flag; if True, PSD will be estimated via coarse-graining
        as opposed to Welch-averaging. Default is False.
    zeropad: bool, optional
        Before doing FFT whether to zero pad the data equal to the length of
        FFT or not. Default is False.
    overlap_factor: float, optional
        Amount of overlap between adjacent segments (range between 0 and 1).
        This factor should be same as the one used for cross_spectral_density.
    window_fftgram_dict: dictionary, optional
        Dictionary containing name and parameters describing which window to
        use for producing FFTs. Default is \"hann\".
    overlap_factor_welch: float
        Overlap factor to use when using Welch's method (NOT coarsegraining). Users should provide proper combination of overlap_factor and window_fftgram_dict. For \"hann\" window use 0.5 overlap_factor and for \"boxcar"\ window use 0 overlap_factor. Default is 0.5 (50% overlap), which is optimal when using Welch's method with a \"hann\" window.

    Returns
    -------
    psd_spectrogram: gwpy PSD spectrogram
        PSD spectrogram with each PSD duration equal to segment duration
    """

    psd_spectrogram = cross_spectral_density(
        time_series_data1=time_series_data,
        time_series_data2=None,
        segment_duration=segment_duration,
        frequency_resolution=frequency_resolution,
        coarse_grain=coarse_grain,
        overlap_factor=overlap_factor,
        overlap_factor_welch=overlap_factor_welch,
        zeropad=zeropad,
        window_fftgram_dict=window_fftgram_dict,
        is_psd=True,
    )
    return psd_spectrogram


def running_mean(data, coarsening_factor=1, axis=-1):
    """
    Compute the running mean of an array, this uses the default axis of numpy
    cumsum.

    Parameters
    ----------
    data: array-like
        Array of size M to be average
    coarsening_factor: int
        Number of segments to average, default=1
    axis:
        Axis to apply the mean over, default=-1

    Returns
    -------
    array-like: the averaged array of size M - coarsening factor
    """
    coarsening_factor = int(coarsening_factor)
    if axis != -1:
        data = np.swapaxes(axis, -1)
    cumsum = np.cumsum(np.insert(data, 0, 0))
    return (
        np.swapaxes(cumsum[coarsening_factor:] - cumsum[:-coarsening_factor], axis, -1)
        / coarsening_factor
    )



import logging
import os

import bilby.gw.detector
from bilby.gw.detector import PowerSpectralDensity
from gwpy.segments import SegmentList

from .preprocessing import (
    preprocessing_data_channel_name,
    preprocessing_data_gwpy_timeseries,
    preprocessing_data_timeseries_array,
    self_gate_data,
)
from .spectral import before_after_average, power_spectral_density


class Interferometer(bilby.gw.detector.Interferometer):

    """Subclass of bilby's Interferometer class"""

    def __init__(self, *args, **kwargs):
        """Instantiate an Interferometer class

        Parameters
        ==========
        *args : arguments passed to the (parent) bilby's Interferometer class.
        **kwargs : keyword arguments passed to the (parent) bilby's Inteferometer class.

        Nominally, the bilby's Interferometer class takes the following arguments.
        name: str
            Interferometer name, e.g. H1.
        power_spectral_density: bilby.gw.detector.PowerSpectralDensity
            Power spectral density determining the sensitivity of the detector.
        minimum_frequency: float
            Minimum frequency to analyse for detector.
        maximum_frequency: float
            Maximum frequency to analyse for detector.
        length: float
            Length of the interferometer in km.
        latitude: float
            Latitude North in degrees (South is negative).
        longitude: float
            Longitude East in degrees (West is negative).
        elevation: float
            Height above surface in metres.
        xarm_azimuth: float
            Orientation of the x arm in degrees North of East.
        yarm_azimuth: float
            Orientation of the y arm in degrees North of East.
        xarm_tilt: float, optional
            Tilt of the x arm in radians above the horizontal defined by
            ellipsoid earth model in LIGO-T980044-08.
        yarm_tilt: float, optional
            Tilt of the y arm in radians above the horizontal.
        calibration_model: Recalibration
            Calibration model, this applies the calibration correction to the
            template, the default model applies no correction.

        See https://lscsoft.docs.ligo.org/bilby/api/bilby.gw.detector.interferometer.Interferometer.html#bilby.gw.detector.interferometer.Interferometer
        for the detailed docs of the parent class.

        Additional attributes
        timeseries : gwpy timeseries
            timeseries object with resampling/high-pass filter applied.
        psd_spectrogram : gwpy spectrogram
            gwpy spectrogram of power spectral density
        gates: gwpy SegmentList
            List of segments that have been gated, not including any additional padding.
        gate_pad: float
            Duration of padding used when applying gates.

        """
        self.gates = SegmentList()
        self.gate_pad = None
        super(Interferometer, self).__init__(*args, **kwargs)

    @classmethod
    def get_empty_interferometer(cls, name):
        """
        A classmethod to get an Interferometer class from a given ifo name

        Parameters
        ==========
        name : str
            Interferometer name, e.g. H1.

        Returns
        =======
        interferometer: Interferometer
            Interferometer instance

        """
        filename = os.path.join(
            os.path.dirname(bilby.gw.detector.__file__),
            "detectors",
            f"{name}.interferometer",
        )
        try:
            parameters = dict()
            with open(filename, "r") as parameter_file:
                lines = parameter_file.readlines()
                for line in lines:
                    if line[0] == "#" or line[0] == "\n":
                        continue
                    split_line = line.split("=")
                    key = split_line[0].strip()
                    value = eval(
                        "=".join(split_line[1:]),
                        dict(__builtins__=dict()),
                        dict(PowerSpectralDensity=PowerSpectralDensity),
                    )
                    parameters[key] = value
            if "shape" not in parameters.keys():
                logging.debug("Assuming L shape for name")
            elif parameters["shape"].lower() in ["l", "ligo"]:
                parameters.pop("shape")
            elif parameters["shape"].lower() in ["triangular", "triangle"]:
                raise ValueError("Triangular detectros are not implemented yet.")
            else:
                raise IOError(
                    f"{filename} could not be loaded. Invalid parameter 'shape'."
                )
            ifo_cls = cls(**parameters)
            ifo_cls._check_ifo_name(name)
            return ifo_cls
        except OSError:
            raise ValueError(f"Interferometer {name} not implemented")

    @classmethod
    def from_parameters(cls, name, parameters):
        """
        A classmethod to get an Interferometer class from a given argparser object

        Parameters
        ==========
        name : str
            Interferometer name, e.g. H1.
        parameters : argparser object
            This contains attributes defined for command line options

        Returns
        =======
        interferometer: Interferometer
            Interferometer instance

        """
        ifo = cls.get_empty_interferometer(name)
        channel = str(ifo.name + ":" + parameters.channel)

        ifo.set_timeseries_from_channel_name(
            channel,
            t0=parameters.t0,
            tf=parameters.tf,
            data_type=parameters.data_type,
            frametype=parameters.frametype,
            local_data_path=parameters.local_data_path,
            new_sample_rate=parameters.new_sample_rate,
            cutoff_frequency=parameters.cutoff_frequency,
            segment_duration=parameters.segment_duration,
            number_cropped_seconds=parameters.number_cropped_seconds,
            window_downsampling=parameters.window_downsampling,
            ftype=parameters.ftype,
            time_shift=parameters.time_shift,
            tag=parameters.tag,
            input_sample_rate=parameters.input_sample_rate,
        )
        return ifo

    def set_timeseries_from_channel_name(self, channel, **kwargs):
        """
        Set a timeseries attribute from a given channel name

        Parameters
        ==========
        channel: str
            Name of the channel (e.g.: "L1:GWOSC-4KHZ_R1_STRAIN") from which to load the data.

        **kwargs : keyword arguments passed to preprocess module.

        """

        t0 = kwargs.pop("t0")
        tf = kwargs.pop("tf")
        data_type = kwargs.pop("data_type")
        frametype = kwargs.pop("frametype")
        local_data_path = kwargs.pop("local_data_path")
        new_sample_rate = kwargs.pop("new_sample_rate")
        input_sample_rate = kwargs.pop("input_sample_rate")
        cutoff_frequency = kwargs.pop("cutoff_frequency")
        segment_duration = kwargs.pop("segment_duration")
        number_cropped_seconds = kwargs.pop("number_cropped_seconds")
        window_downsampling = kwargs.pop("window_downsampling")
        ftype = kwargs.pop("ftype")
        time_shift = kwargs.pop("time_shift")
        tag = kwargs.pop("tag")
        self.duration = segment_duration
        self.timeseries = preprocessing_data_channel_name(
            IFO=self.name,
            channel=channel,
            t0=t0,
            tf=tf,
            data_type=data_type,
            frametype=frametype,
            local_data_path=local_data_path,
            new_sample_rate=new_sample_rate,
            cutoff_frequency=cutoff_frequency,
            segment_duration=segment_duration,
            number_cropped_seconds=number_cropped_seconds,
            window_downsampling=window_downsampling,
            ftype=ftype,
            time_shift=time_shift,
            tag=tag,
            input_sample_rate=input_sample_rate,
        )
        self._check_timeseries_channel_name(channel)
        self.sampling_frequency = new_sample_rate

    def set_timeseries_from_timeseries_array(
        self, timeseries_array, sample_rate, **kwargs
    ):
        """
        Set a timeseries attribute from a given numpy array

        Parameters
        ==========
        timeseries_array: numpy array
            timeseries strain data as numpy array object

        **kwargs : keyword arguments passed to preprocess module.

        """

        t0 = kwargs.pop("t0")
        tf = kwargs.pop("tf")
        data_type = kwargs.pop("data_type")
        new_sample_rate = kwargs.pop("new_sample_rate")
        cutoff_frequency = kwargs.pop("cutoff_frequency")
        segment_duration = kwargs.pop("segment_duration")
        number_cropped_seconds = kwargs.pop("number_cropped_seconds")
        window_downsampling = kwargs.pop("window_downsampling")
        ftype = kwargs.pop("ftype")
        time_shift = kwargs.pop("time_shift")
        self.duration = segment_duration
        self.timeseries = preprocessing_data_timeseries_array(
            IFO=self.name,
            array=timeseries_array,
            t0=t0,
            tf=tf,
            new_sample_rate=new_sample_rate,
            cutoff_frequency=cutoff_frequency,
            segment_duration=segment_duration,
            sample_rate=sample_rate,
            number_cropped_seconds=number_cropped_seconds,
            window_downsampling=window_downsampling,
            ftype=ftype,
            time_shift=time_shift,
        )
        self.timeseries.channel = kwargs.pop("channel")
        self._check_timeseries_sample_rate(new_sample_rate)
        self.sampling_frequency = sample_rate

    def set_timeseries_from_gwpy_timeseries(self, gwpy_timeseries, **kwargs):
        """
        Set a timeseries attribute from a given gwpy timeseries object

        Parameters
        ==========
        gwpy_timeseries: gwpy.timeseries
            timeseries strain data as gwpy.timeseries object

        **kwargs : keyword arguments passed to preprocess module.

        """

        new_sample_rate = kwargs.pop("new_sample_rate")
        segment_duration = kwargs.pop("segment_duration")
        cutoff_frequency = kwargs.pop("cutoff_frequency")
        number_cropped_seconds = kwargs.pop("number_cropped_seconds")
        window_downsampling = kwargs.pop("window_downsampling")
        ftype = kwargs.pop("ftype")
        time_shift = kwargs.pop("time_shift")
        self.duration = segment_duration
        self.timeseries = preprocessing_data_gwpy_timeseries(
            IFO=self.name,
            gwpy_timeseries=gwpy_timeseries,
            new_sample_rate=new_sample_rate,
            cutoff_frequency=cutoff_frequency,
            number_cropped_seconds=number_cropped_seconds,
            window_downsampling=window_downsampling,
            ftype=ftype,
            time_shift=time_shift,
        )
        self.timeseries.channel = kwargs.pop("channel")
        self._check_timeseries_sample_rate(new_sample_rate)
        self.sampling_frequency = new_sample_rate

    def set_psd_spectrogram(
        self,
        frequency_resolution,
        coarse_grain=False,
        overlap_factor=0.5,
        window_fftgram_dict={"window_fftgram": "hann"},
        overlap_factor_welch=0.5,
    ):
        """
        Set psd_spectrogram attribute from a given spectrum-related information.

        Parameters
        ==========
        frequency_resolution: float
            Frequency resolution of the final PSDs; This sets the time duration
            over which FFTs are calculated in the pwelch method
        coarse_grain: bool
            Coarse-graining flag; if True, PSD will be estimated via coarse-graining
            as opposed to Welch-averaging. Default is False.
        overlap_factor: float, optional
            Amount of overlap between adjacent segments (range between 0 and 1)
            This factor should be same as the one used for cross_spectral_density
            (default 0, no overlap)
        window_fftgram_dict: dictionary, optional
            Dictionary containing name and parameters describing which window to use when producing fftgrams for psd estimation. Default is \"hann\".
        overlap_factor_welch: float, optional
            Overlap factor to use when if using Welch's method to estimate the PSD (NOT coarsegraining). For \"hann\" window use 0.5 overlap_factor and for \"boxcar"\ window use 0 overlap_factor. Default is 0.5 (50% overlap), which is optimal when using Welch's method with a \"hann\" window.

        """

        # PSD estimation needs zeropadding when using coarse-graining
        zeropad_psd = coarse_grain

        self.psd_spectrogram = power_spectral_density(
            self.timeseries,
            self.duration,
            frequency_resolution,
            coarse_grain=coarse_grain,
            zeropad=zeropad_psd,
            overlap_factor=overlap_factor,
            window_fftgram_dict=window_fftgram_dict,
            overlap_factor_welch=overlap_factor_welch,
        )
        self.psd_spectrogram.channel = self.timeseries.channel
        self._check_spectrogram_channel_name(self.timeseries.channel.name)
        self._check_spectrogram_frequency_resolution(frequency_resolution)

    def set_average_psd(self, N_average_segments=2):
        """
        Set average_psd attribute from the existing raw psd

        Parameters
        ==========
        N_average_segments: int
            Number of segments used for PSD averaging (from both sides of the segment of interest)
            N_avg_segs should be even and >= 2

        """
        try:
            self.average_psd = before_after_average(
                self.psd_spectrogram, self.duration, N_average_segments
            )
        except AttributeError:
            print(
                "PSDs have not been calculated yet! Need to set_psd_spectrogram first."
            )

    def gate_data_apply(self, **kwargs):
        """
        Self-gate the tgwpy timeseries associated with this timeseries. The list
        of times gated and the padding applied are stored as properties of the Interferometer.

        Parameters
        ==========
        gate_tzero : float
            half-width time duration (seconds) in which the timeseries is
            set to zero
        gate_tpad : float
            half-width time duration (seconds) in which the Planck window
            is tapered
        gate_threshold : float
            amplitude threshold, if the data exceeds this value a gating window
            will be placed
        cluster_window : float
            time duration (seconds) over which gating points will be clustered
        gate_whiten : bool
            if True, data will be whitened before gating points are discovered,
            use of this option is highly recommended

        """
        gate_tzero = kwargs.pop("gate_tzero")
        gate_tpad = kwargs.pop("gate_tpad")
        gate_threshold = kwargs.pop("gate_threshold")
        cluster_window = kwargs.pop("cluster_window")
        gate_whiten = kwargs.pop("gate_whiten")
        self.timeseries, new_gates = self_gate_data(
            self.timeseries,
            tzero=gate_tzero,
            tpad=gate_tpad,
            gate_threshold=gate_threshold,
            cluster_window=cluster_window,
            whiten=gate_whiten,
        )
        self.gates = self.gates | new_gates
        self.gate_pad = gate_tpad

    def apply_gates_from_file(self, loaded_object, index, **kwargs):
        """
        Load gates from a pygwb output file and apply them to the Interferometer object. 
        The gated times are stored as a property of the object.
        
        Parameters
        ==========
        loaded_object : 
            Object that represents the data in the output file, e.g. a loaded npzobject.
        index ; int
            Integer representing the correct ifo object in the baseline
        gate_tpad : float
            half-width time duration (seconds) in which the Planck window
            is tapered
        """
        gates = loaded_object[f"ifo_{index}_gates"]
        gate_tpad = kwargs.pop("gate_tpad")

        self.timeseries, new_gates = self_gate_data(
            self.timeseries,
            tpad=gate_tpad,
            gates=gates,
        )

        self.gates = self.gates | new_gates
        self.gate_pad = gate_tpad

    def _check_ifo_name(self, name):
        if not self.name == name:
            raise AssertionError(
                "The ifo name in Interferomaeter class does not match given name!"
            )

    def _check_timeseries_channel_name(self, channel_name):
        if not self.timeseries.channel.name == channel_name:
            raise AssertionError(
                "Channel name in timeseries does not match given channel!"
            )

    def _check_timeseries_sample_rate(self, sample_rate):
        if not self.timeseries.sample_rate.value == sample_rate:
            raise AssertionError(
                "Sampling rate in timeseries does not match given sampleing rate!"
            )

    def _check_spectrogram_channel_name(self, channel_name):
        if not self.psd_spectrogram.channel.name == channel_name:
            raise AssertionError(
                "Channel name in psd_spectrogram does not match given channel!"
            )

    def _check_spectrogram_frequency_resolution(self, frequency_resolution):
        if not self.psd_spectrogram.df.value == frequency_resolution:
            raise AssertionError(
                "Frequency resolution in psd_spectrogram does not match given frequency resolution!"
            )



import copy
import os
import warnings

import lal
import numpy as np
import scipy
from gwpy import timeseries
from gwpy.segments import Segment, SegmentList
from gwsumm.data.timeseries import get_timeseries


def set_start_time(
    job_start_GPS: int,
    job_end_GPS: int,
    buffer_secs: int,
    segment_duration: int,
    do_sidereal: bool = False,
):
    """
    Function to identify segment start times
    either with or without sidereal option

    Parameters
    ==========
    job_start_GPS: int
        Integer indicating the start time (in GPS)
        of the data

    job_end_GPS: int
        Integer indicating the end time (in GPS)
        of the data

    buffer_secs: int
        Number of cropped seconds

    segment_duration: int
        Duration of each segment

    do_sidereal: bool
        ---

    Returns
    =======
    centered_start_time: int
        Integer with the initial time of the segment to be pre-processed
    """
    if not do_sidereal:
        job_duration = job_end_GPS - job_start_GPS
        M = np.floor((job_duration - 2 * buffer_secs) / segment_duration)
        centered_start_time = (
            job_start_GPS
            + buffer_secs
            + np.floor((job_duration - 2 * buffer_secs - M * segment_duration) / 2)
        )
    else:
        srfac = 23.9344696 / 24
        # sidereal time conversion factor
        srtime = (lal.GreenwichMeanSiderealTime(job_end_GPS) % (2 * np.pi)) * 3600
        md = np.mod(srtime, segment_duration / srfac)
        centered_start_time = np.round(job_end_GPS + segment_duration - md * srfac)
        if centered_start_time - job_end_GPS < buffer_secs:
            centered_start_time = centered_start_time + segment_duration
    return centered_start_time


def read_data(
    IFO: str,
    data_type: str,
    channel: str,
    t0: int,
    tf: int,
    local_data_path: str = "",
    frametype: str = "",
    tag: str = "C00",
    input_sample_rate: int = 16384,
):
    """
    Function doing the reading of the data to be used in the
    stochastic pipeline

    Parameters
    ==========
    IFO: string
        Interferometer from which to retrieve the data

    data_type: string
        String indicating the type of data to be read,
        either:
        - 'public' : data from GWOSC (https://www.gw-openscience.org/)
        - 'private' : data from the LIGO-Virgo servers restricted to members of the collaboration
        - 'local' (if 'local_data_path'): data the user has saved locally

    channel: string
        Name of the channel (e.g.: "L1:GWOSC-4KHZ_R1_STRAIN")

    t0: int
        GPS time of the start of the data taking

    tf: int
        GPS time of the end of the data taking

    frametype: string
        Frame type that contains the channel, only used if data_type=private (e.g.: "L1_HOFT_C00")

    local_data_path: str, optional
        path where local gwf is stored

    tag: str, optional
        tag identifying type of data, e.g.: 'C00', 'C01'

    input_sample_rate: int
        Sampling rate of the timeseries to be read

    Returns
    =======
    data: TimeSeries object
        Time series containing the requested data
    """
    if data_type == "public":
        data = timeseries.TimeSeries.fetch_open_data(
            IFO, t0, tf, sample_rate=input_sample_rate
        )
        data.channel = channel
    elif data_type == "private":
        if frametype=="":
            frametype=None
        data = get_timeseries(channel, segments=[[t0, tf]], frametype=frametype)
        if len(data) > 1:
            raise ValueError("Something went wrong while getting the data!"
                             "There was more than one data stretch returned.")
        else:
            data = data[0]
        data.channel = channel
    elif data_type == "local":
        if os.path.isdir(local_data_path):
            local_data = []
            for f in os.listdir(local_data_path):
                local_data.append(os.path.join(local_data_path, f))
        else:
            local_data = local_data_path
        data = timeseries.TimeSeries.read(
            source=local_data, channel=channel, start=t0, end=tf, verbose=True
        )
        data.channel = channel
        data.name = IFO

    else:
        raise ValueError("Wrong data type. Choose between: public, private and local")
    return data


def apply_high_pass_filter(
    timeseries: timeseries.TimeSeries,
    sample_rate: int,
    cutoff_frequency: float,
    number_cropped_seconds: int = 2,
):
    """
    Function to apply a high pass filter to a timeseries

    Parameters
    ==========
    timeseries: gwpy_timeseries
        Timeseries to which to apply the high pass filter

    sample_rate: int
        Sampling rate of the timeseries

    cutoff_frequency: float
        Frequency (in Hz) from which to start applying the
        high pass filter

    number_cropped_seconds: int, optional
        Number of seconds to remove at the beginning and end
        of the high-passed data; default is 2

    Returns
    =======
    filtered: Timeseries
        High-pass filtered timeseries
    """
    zpk = scipy.signal.butter(
        16, cutoff_frequency, "high", analog=False, output="zpk", fs=sample_rate
    )
    filtered = timeseries.filter(zpk, filtfilt=True)
    filtered = filtered.crop(*filtered.span.contract(number_cropped_seconds))
    return filtered


def resample_filter(
    time_series_data: timeseries.TimeSeries,
    new_sample_rate: int,
    cutoff_frequency: float,
    number_cropped_seconds: int = 2,
    window_downsampling: str = "hamming",
    ftype: str = "fir",
):
    """
    Function doing part of the pre-processing
    (resampling,filtering and fftgram computation)
    of the data to be used in the stochastic pipeline

    Parameters
    ==========

    time_series_data: gwpy_timeseries
        timeseries data to be analysed in the pipeline

    new_sample_rate: int
        Sampling rate of the downsampled-timeseries

    cutoff_frequency: float
        Frequency (in Hz) from which to start applying the
        high pass filter

    number_cropped_seconds: int
        Number of seconds to remove at the beginning and end
        of the high-passed data

    window_downsampling: string
        Type of window used to downsample

    ftype: string
        Type of filter to use in the downsampling


    Returns
    =======
    filtered: gwpy_timeseries
        Timeseries containing the filtered and high passed data
    """
    if (new_sample_rate * number_cropped_seconds) < 18:
        warnings.warn(
            f"Number of cropped seconds requested {number_cropped_seconds}s is low compared to the sampling rate {new_sample_rate}: cropped-seconds x sampling-rate = {number_cropped_seconds*new_sample_rate}."
        )
    data_to_resample = copy.deepcopy(time_series_data.value)
    original_times = copy.deepcopy(time_series_data.times)
    nan_mask = np.isnan(time_series_data.value)  # .flatten()

    if np.sum(nan_mask) != 0:
        data_nansafe = data_to_resample[~nan_mask]
        times_nansafe = original_times[~nan_mask]
        interped_data = scipy.interpolate.CubicSpline(times_nansafe, data_nansafe)
        new = interped_data(original_times).view(time_series_data.__class__)
        new.__metadata_finalize__(time_series_data)
        new._unit = time_series_data.unit
        resampled = new.resample(new_sample_rate, window_downsampling, ftype)
        warnings.warn(
            f"There are {np.sum(nan_mask)} NaNs in the timestream ({np.sum(nan_mask)*100/len(time_series_data)}% of the data). These will be ignored in pre-processing."
        )
    else:
        resampled = time_series_data.resample(
            new_sample_rate, window_downsampling, ftype
        )

    sample_rate = resampled.sample_rate.value
    filtered = apply_high_pass_filter(
        timeseries=resampled,
        sample_rate=sample_rate,
        cutoff_frequency=cutoff_frequency,
        number_cropped_seconds=number_cropped_seconds,
    )

    return filtered


def self_gate_data(
    time_series_data: timeseries.TimeSeries,
    tzero: float = 1.0,
    tpad: float = 0.5,
    gate_threshold: float = 50.0,
    cluster_window: float = 0.5,
    whiten: bool = True,
    gates: SegmentList = None
):
    """
    Function to self-gate
    data to be used in the stochastic pipeline

    Parameters
    ==========

    time_series_data: gwpy_timeseries
        timeseries data to be analysed in the pipeline

    tzero : `int`, optional
        half-width time duration (seconds) in which the timeseries is
        set to zero

    tpad : `int`, optional
        half-width time duration (seconds) in which the Planck window
        is tapered

    whiten : `bool`, optional
        if True, data will be whitened before gating points are discovered,
        use of this option is highly recommended

    threshold : `float`, optional
        amplitude threshold, if the data exceeds this value a gating window
        will be placed

    cluster_window : `float`, optional
        time duration (seconds) over which gating points will be clustered

    Returns
    =======
    gated: gwpy_timeseries
        Timeseries containing the gated data

    deadtime: `gwpy.segments.SegmentList`
        SegmentList containing the times that were gated, not including
        any padding applied

    Notes
    -----
    This method is based on `gwpy.timeseries.gate`. See
    https://gwpy.github.io/docs/latest/api/gwpy.timeseries.TimeSeries/?highlight=timeseries#gwpy.timeseries.TimeSeries.gate
    for additional details.
    """
    from scipy.signal import find_peaks

    # Find points to gate based on a threshold
    sample = time_series_data.sample_rate.to("Hz").value
    if gates is None:
        data = time_series_data.whiten() if whiten else time_series_data
        window_samples = cluster_window * sample
        gates = find_peaks(abs(data.value), height=gate_threshold, distance=window_samples)[
            0
        ]
    # represent gates as time segments
        deadtime = SegmentList(
            [
                Segment(
                    time_series_data.t0.value + (k / sample) - tzero,
                    time_series_data.t0.value + (k / sample) + tzero,
                )
                for k in gates
            ]
        ).coalesce()
    else:
        deadtime = SegmentList([Segment(k[0], k[1]) for k in gates]).coalesce()

    # return the self-gated timeseries
    gated = time_series_data.mask(deadtime=deadtime, const=0, tpad=tpad)
    return gated, deadtime


def shift_timeseries(time_series_data: timeseries.TimeSeries, time_shift: int = 0):

    """
    Function that shifts a timeseries by an amount time_shift
    in order to perform the time shifted analysis

    Parameters
    ==========

    time_series_data: gwpy_timeseries
        timeseries data to be analysed in the pipeline

    time_shift: int
        value of the time shift (in seconds)

    Returns
    =======
    shifted_data: gwpy_timeseries
        Timeseries containing the shifted_data
    """

    if time_shift > 0:
        shifted_data = np.roll(
            time_series_data, int(time_shift / time_series_data.dt.value)
        )
    else:
        shifted_data = time_series_data
    return shifted_data


def preprocessing_data_gwpy_timeseries(
    IFO: str,
    gwpy_timeseries: timeseries.TimeSeries,
    new_sample_rate: int,
    cutoff_frequency: float,
    number_cropped_seconds: int = 2,
    window_downsampling: str = "hamming",
    ftype: str = "fir",
    time_shift: int = 0,
):
    """
    Function doing the pre-processing of a gwpy timeseries to be used in the
    stochastic pipeline

    Parameters
    ==========

    IFO: string
        Interferometer from which to retrieve the data

    gwpy_timeseries: gwpy_timeseries
        Timeseries from gwpy

    new_sample_rate:int
        Sampling rate of the downsampled-timeseries

    cutoff_frequency: float
        Frequency (in Hz) from which to start applying the
        high pass filter

    number_cropped_seconds: int
        Number of seconds to remove at the beginning and end
        of the high-passed data

    window_downsampling: string
        Type of window used to downsample

    ftype: string
        Type of filter to use in the downsampling

    time_shift: int
        value of the time shift (in seconds)

    Returns
    =======
    pre_processed_data: gwpy_timeseries
        Timeseries containing the filtered and high passed data (shifted if time_shift>0)
    """
    time_series_data = gwpy_timeseries
    filtered = resample_filter(
        time_series_data=time_series_data,
        new_sample_rate=new_sample_rate,
        cutoff_frequency=cutoff_frequency,
        number_cropped_seconds=number_cropped_seconds,
        window_downsampling=window_downsampling,
        ftype=ftype,
    )
    if time_shift > 0:
        return shift_timeseries(time_series_data=filtered, time_shift=time_shift)
    else:
        return filtered


def preprocessing_data_channel_name(
    IFO: str,
    t0: int,
    tf: int,
    data_type: str,
    channel: str,
    new_sample_rate: int,
    cutoff_frequency: float,
    segment_duration: int,
    number_cropped_seconds: int = 2,
    window_downsampling: str = "hamming",
    ftype: str = "fir",
    time_shift: int = 0,
    local_data_path: str = "",
    frametype: str = "",
    tag: str = "C00",
    input_sample_rate: int = 16384,
):
    """
    Function doing the pre-processing of the data to be used in the
    stochastic pipeline

    Parameters
    ==========
    IFO: string
        Interferometer from which to retrieve the data

    t0: int
        GPS time of the start of the data taking

    tf: int
        GPS time of the end of the data taking

    data_type: string
        String indicating the type of data to be read,
        either 'public' or 'private'

    channel: string
        Name of the channel (e.g.: "L1:GWOSC-4KHZ_R1_STRAIN")

    frametype: string
       Frame type that contains the channel, only used if data_type=private (e.g.: "L1_HOFT_C00") 

    new_sample_rate:int
        Sampling rate of the downsampled-timeseries

    cutoff_frequency: float
        Frequency (in Hz) from which to start applying the
        high pass filter

    segment_duration: int
        Duration of each segment (argument of set_start_time)

    number_cropped_seconds: int
        Number of seconds to remove at the beginning and end
        of the high-passed data

    window_downsampling: string
        Type of window used to downsample

    ftype: string
        Type of filter to use in the downsampling

    time_shift: int
        value of the time shift (in seconds)

    tag: str, optional
        tag identifying type of data, e.g.: 'C00', 'C01'

    input_sample_rate: int
        Sampling rate of the timeseries to be read

    Returns
    =======
    pre_processed_data: gwpy_timeseries
        Timeseries containing the filtered and high passed data (shifted if time_shift>0)
    """
    data_start_time = set_start_time(
        job_start_GPS=t0,
        job_end_GPS=tf,
        buffer_secs=number_cropped_seconds,
        segment_duration=segment_duration,
    )

    time_series_data = read_data(
        IFO=IFO,
        data_type=data_type,
        channel=channel,
        frametype=frametype,
        t0=data_start_time - number_cropped_seconds,
        tf=tf,
        local_data_path=local_data_path,
        tag=tag,
        input_sample_rate=input_sample_rate,
    )

    return preprocessing_data_gwpy_timeseries(
        IFO=IFO,
        gwpy_timeseries=time_series_data,
        new_sample_rate=new_sample_rate,
        cutoff_frequency=cutoff_frequency,
        number_cropped_seconds=number_cropped_seconds,
        window_downsampling=window_downsampling,
        ftype=ftype,
        time_shift=time_shift,
    )


def preprocessing_data_timeseries_array(
    t0: int,
    tf: int,
    IFO: str,
    array: np.ndarray,
    new_sample_rate: int,
    cutoff_frequency: float,
    segment_duration: int,
    sample_rate: int = 4096,
    number_cropped_seconds: int = 2,
    window_downsampling: str = "hamming",
    ftype: str = "fir",
    time_shift: int = 0,
):
    """
    Function doing the pre-processing of a time-series array to be used in the
    stochastic pipeline

    Parameters
    ==========
    t0: int
        GPS time of the start of the data taking

    tf: int
        GPS time of the end of the data taking

    IFO: string
        Interferometer from which to retrieve the data

    array: array
        Array containing a timeseries

    new_sample_rate:int
        Sampling rate of the downsampled-timeseries

    cutoff_frequency: float
        Frequency (in Hz) from which to start applying the
        high pass filter

    segment_duration: int
        Duration of each segment (argument of set_start_time)

    sample_rate: int
        Sampling rate of the original timeseries

    number_cropped_seconds: int
        Number of seconds to remove at the beginning and end
        of the high-passed data

    window_downsampling: string
        Type of window used to downsample

    ftype: string
        Type of filter to use in the downsampling

    time_shift: int
        value of the time shift (in seconds)

    Returns
    =======
    pre_processed_data: gwpy_timeseries
        Timeseries containing the filtered and high passed data (shifted if time_shift>0)
    """
    data_start_time = set_start_time(
        job_start_GPS=t0,
        job_end_GPS=tf,
        buffer_secs=number_cropped_seconds,
        segment_duration=segment_duration,
    )
    time_series_data = timeseries.TimeSeries(
        array, t0=data_start_time, sample_rate=sample_rate
    )
    return preprocessing_data_gwpy_timeseries(
        IFO=IFO,
        gwpy_timeseries=time_series_data,
        new_sample_rate=new_sample_rate,
        cutoff_frequency=cutoff_frequency,
        number_cropped_seconds=number_cropped_seconds,
        window_downsampling=window_downsampling,
        ftype=ftype,
        time_shift=time_shift,
    )



import numpy as np
import scipy.ndimage as ndi
from loguru import logger

from pygwb.constants import H0
from pygwb.omega_spectra import OmegaSpectrum

from .util import _check_omegaspectra, calc_bias, window_factors


def postprocess_Y_sigma(
    Y_fs,
    var_fs,
    segment_duration,
    deltaF,
    new_sample_rate,
    frequency_mask=True,
    badtimes_mask=None,
    window_fftgram_dict={"window_fftgram": "hann"},
    window_fftgram_dict_welch={"window_fftgram": "hann"},
    overlap_factor=0.5,
    overlap_factor_welch=0.5,
    N_avg_segs=2,
):
    """Run postprocessing of point estimate and sigma spectrograms, combining even and
    odd segments in the case of overlapping data. For more details see - https://dcc.ligo.org/public/0027/T040089/000/T040089-00.pdf

    Parameters:
    -----------
    Y_fs : array-like
        2D array of point estimates with Ntimes x Nfreqs with overlapping segments
    var_fs : array-like
        2D array of variances or 2D with dimensions Ntimes x Nfreqs with overlapping time segments
    segment_duration : float
        Duration of each time segment
    deltaF : float
        Frequency resolution
    new_sample_rate : float
        sample rate of timeseries after resampling
    frequency_mask: array-like, optional
        Boolean mask to apply to frequencies for the calculation. Defaults to True which includes all frequencies in the analysis.
    badtimes_mask: array-like, optional
        Boolean mask to apply to GPStimes in the calculation. Defaults to None such that all times are included.
    window_fftgram_dict: dictionary, optional
        Dictionary with window characteristics used in PSD estimation. Default is `(window_fftgram_dict={"window_fftgram": "hann"}`
    overlap_factor: float, optional
        Overlap factor used in PSD estimation. Default is 0.5.

    Returns:
    --------
    Y_f_new : array-like
        1D point estimate spectrum
    sigma_f_few : array-like
        1D sigma spectrum
    """
    if badtimes_mask is None:
        badtimes_mask = np.zeros(len(Y_fs), dtype=bool)

    goodtimes_mask = ~badtimes_mask
    labels, n_labels = ndi.label(goodtimes_mask)

    Y_fs_sliced = []
    var_fs_sliced = []

    for sli in ndi.find_objects(labels):
        Y = Y_fs[sli]
        var = var_fs[sli]

        if len(Y) == 1:
            Y_fs_sliced.append(Y[0])
            var_fs_sliced.append(var[0])
        else:
            Y_red, var_red = odd_even_segment_postprocessing(
                Y,
                var,
                segment_duration,
                new_sample_rate,
                frequency_mask=frequency_mask,
                window_fftgram_dict=window_fftgram_dict,
                overlap_factor=overlap_factor,
                N_avg_segs=N_avg_segs,
            )
            Y_fs_sliced.append(Y_red)
            var_fs_sliced.append(var_red)

    Y_fs_sliced = np.array(Y_fs_sliced)
    var_fs_sliced = np.array(var_fs_sliced)

    Y_f_new, sigma_f_new = combine_spectra_with_sigma_weights(
        Y_fs_sliced, np.sqrt(var_fs_sliced)
    )

    bias = calc_bias(
        segment_duration,
        deltaF,
        1 / new_sample_rate,
        N_avg_segs=N_avg_segs,
        window_fftgram_dict=window_fftgram_dict_welch,
        overlap_factor=overlap_factor_welch,
    )
    logger.debug(f"Bias factor: {bias}")
    sigma_f_new *= bias

    return Y_f_new, sigma_f_new


def odd_even_segment_postprocessing(
    Y_fs,
    var_fs,
    segment_duration,
    new_sample_rate,
    frequency_mask=True,
    window_fftgram_dict={"window_fftgram": "hann"},
    overlap_factor=0.5,
    N_avg_segs=2,
):
    """Perform averaging which combines even and odd segments for overlapping data. 

    Parameters:
    -----------
    Y_fs : array-like
        2D array of point estimates with Ntimes x Nfreqs with overlapping segments
    var_fs : array-like
        2D array of variances or 2D with dimensions Ntimes x Nfreqs with overlapping time segments
    segment_duration : float
        Duration of each time segment
    new_sample_rate : float
        sample rate of timeseries after resampling
    frequency_mask: array-like, optional
        Boolean mask to apply to frequencies for the calculation
    window_fftgram_dict: dictionary, optional
        Dictionary with window characteristics used in PSD estimation. Default is `(window_fftgram_dict={"window_fftgram": "hann"}`
    overlap_factor: float, optional

    Returns:
    --------
    Y_f_new : array-like
        1D point estimate spectrum
    var_f_few : array-like
        1D sigma spectrum
    """
    _, w1w2squaredbar, _, w1w2squaredovlbar = window_factors(
        int(segment_duration * new_sample_rate), window_fftgram_dict, overlap_factor=overlap_factor
    )
    k = w1w2squaredovlbar / w1w2squaredbar

    size = np.size(Y_fs, axis=0)
    # even/odd indices
    evens = np.arange(0, size, 2)
    odds = np.arange(1, size, 2)

    X_even = np.nansum(Y_fs[evens] / var_fs[evens], axis=0)
    GAMMA_even = np.nansum(var_fs[evens] ** -1, axis=0)
    X_odd = np.nansum(Y_fs[odds] / var_fs[odds], axis=0)
    GAMMA_odd = np.nansum(var_fs[odds] ** -1, axis=0)

    sigma2_oo = 1 / np.nansum(GAMMA_odd[frequency_mask])
    sigma2_ee = 1 / np.nansum(GAMMA_even[frequency_mask])
    sigma2_1 = 1 / np.nansum(var_fs[0, frequency_mask] ** -1)
    sigma2_N = 1 / np.nansum(var_fs[-1, frequency_mask] ** -1)
    sigma2IJ = 1 / sigma2_oo + 1 / sigma2_ee - (1 / 2) * (1 / sigma2_1 + 1 / sigma2_N)

    sigma2_oo, sigma2_ee, sigma2_1, sigma2_N = [
        s if s != np.inf else 0 for s in (sigma2_oo, sigma2_ee, sigma2_1, sigma2_N)
    ]

    Y_f_new = (
        X_odd * (1 - (k / 2) * sigma2_oo * sigma2IJ)
        + X_even * (1 - (k / 2) * sigma2_ee * sigma2IJ)
    ) / (
        GAMMA_even
        + GAMMA_odd
        - k
        * (GAMMA_even + GAMMA_odd - (1 / 2) * (1 / var_fs[0, :] + 1 / var_fs[-1, :]))
    )

    inv_var_f_new = (
        GAMMA_odd
        + GAMMA_even
        - k
        * (GAMMA_odd + GAMMA_even - (1 / 2) * (1 / var_fs[0, :] + 1 / var_fs[-1, :]))
    ) / (1 - (k ** 2 / 4) * sigma2_oo * sigma2_ee * sigma2IJ ** 2)

    var_f_new = 1 / inv_var_f_new

    return Y_f_new, var_f_new


def calc_Y_sigma_from_Yf_sigmaf(
    Y_f, sigma_f, frequency_mask=True, alpha=None, fref=None
):
    """
    Calculate the omega point estimate and sigma from their respective spectra,
    or spectrograms, taking into account the desired spectral weighting.
    To apply weighting, the frequency array associated to the spectra must be supplied.

    If applied to a 1D array, you get single numbers out. If applied to a 2D array, it combines
    over the second dimension. That is, if dimension is Ntimes x Nfrequencies, then the resulting
    spectra are Ntimes long.

    Parameters
    ==========
    Y_f: `pygwb.omega_spectrogram.OmegaSpectrogram`
        Point estimate spectrum
    sigma_f: `pygwb.omega_spectrogram.OmegaSpectrogram`
        Sigma spectrum
    frequency_mask: array-like, optional
        Boolean mask to apply to frequencies for the calculation.
    alpha: float, optional
        Spectral index to use in case re-weighting is requested.
    fref: float, optional
        Reference frequency to use in case re-weighting is requested.

    Returns:
    --------
    Y : array-like or float
        Point estimate or Point estimate spectrum
    sigma : array-like or float
        point estimate standard deviation (theoretical) or spectrum of point estimate
        standard deviations
    Note
    ====
    If passing in spectrograms, the point estimate and sigma will be calculated per
    spectrum, without any time-averaging applied.
    Y_f and sigma_f can also be gwpy.Spectrogram objects, or numpy arrays. In these cases
    however the reweight functionality is not supported.

    """
    # Reweight in case one wants to pass it.
    if alpha is not None or fref is not None:
        Y_f.reweight(new_alpha=alpha, new_fref=fref)
        sigma_f.reweight(new_alpha=alpha, new_fref=fref)

    # now just strip off what we need...
    try:
        Y_f = np.real(Y_f.value)
        var_f = sigma_f.value ** 2
    except AttributeError:
        Y_f = np.real(Y_f)
        var_f = sigma_f ** 2

    if isinstance(frequency_mask, np.ndarray):
        pass
    elif frequency_mask == True:
        if len(Y_f.shape) == 1:
            frequency_mask = np.ones(Y_f.shape[0], dtype=bool)
        elif len(Y_f.shape) == 2:
            frequency_mask = np.ones(Y_f.shape[1], dtype=bool)
    if len(Y_f.shape) == 1 or Y_f.shape[0] == 1:
        if Y_f.shape[0] == 1:
            Y_f = Y_f[0]
            var_f = var_f[0]
        var = 1 / np.sum(var_f[frequency_mask] ** (-1), axis=-1).squeeze()
        Y = np.nansum(Y_f[frequency_mask] * (var / var_f[frequency_mask]), axis=-1)
    # need to make this nan-safe
    elif len(Y_f.shape) == 2:
        var = 1 / np.sum(var_f[:, frequency_mask] ** (-1), axis=-1).squeeze()
        Y = np.einsum(
            "tf, t -> t", Y_f[:, frequency_mask] / var_f[:, frequency_mask], var
        )
    else:
        raise ValueError("The input is neither a spectrum nor a spectrogram.")

    sigma = np.sqrt(var)

    return Y, sigma


def calculate_point_estimate_sigma_spectra(
    freqs,
    csd,
    avg_psd_1,
    avg_psd_2,
    orf,
    sample_rate,
    segment_duration,
    window_fftgram_dict={"window_fftgram": "hann"},
    overlap_factor=0.5,
    fref=25.0,
    alpha=0.0,
):
    """
    Calculate the Omega point estimate and associated sigma integrand,
    given a set of cross-spectral and power-spectral density spectrograms.
    This is particularly useful for statistical checks.

    If CSD is set to None, only returns variance.

    Parameters
    ==========
    freqs: array_like
        Frequencies associated to the spectrograms.
    csd: gwpy Spectrogram
        CSD spectrogram for detectors 1 and 2.
    avg_psd_1: gwpy Spectrogram
        Spectrogram of averaged PSDs for detector 1.
    avg_psd_2: gwpy Spectrogram
        Spectrogram of averaged PSDs for detector 2.
    orf: array_like
        Overlap reduction function.
    sample_rate: float
        Sampling rate of the data.
    segment_duration: float
        Duration of each segment in seconds.
    window_fftgram_dict: dictionary, optional
        Dictionary with window characteristics used in analysis segment estimation. Default is `(window_fftgram_dict={"window_fftgram": "hann"}`
    overlap_factor: float, optional
        Overlap factor used in analysis segment estimation. Default is 0.5.
    fref: float, optional
        Reference frequency to use in the weighting calculation.
        Final result refers to this frequency.
    alpha: float, optional
        Spectral index to use in the weighting.
    """
    S_alpha = 3 * H0.si.value ** 2 / (10 * np.pi ** 2) / freqs ** 3
    S_alpha *= (freqs / fref) ** float(alpha)

    var_fs = (
        1
        / (2 * segment_duration * (freqs[1] - freqs[0]))
        * avg_psd_1
        * avg_psd_2
        / (orf ** 2 * S_alpha ** 2)
    )

    w1w2bar, w1w2squaredbar, _, _ = window_factors(
        int(sample_rate * segment_duration), window_fftgram_dict=window_fftgram_dict, overlap_factor=overlap_factor
    )
    var_fs = var_fs * w1w2squaredbar / w1w2bar ** 2
    if csd is not None:
        Y_fs = (csd) / (orf * S_alpha)
        return Y_fs, var_fs
    else:
        return var_fs


def combine_spectra_with_sigma_weights(main_spectra, weights_spectra):
    r"""
    Combine different statistically independent spectra :math:`S_i(f)` using spectral weights :math:`w_i(f)`, as


    .. math::

        S(f) = \frac{\sum_i \frac{S_i(f)}{w^2_i(f)}}{\sum_i \frac{1}{w^2_i(f)}},\,\,\,\, \sigma = \sqrt{\frac{1}{\sum_i \frac{1}{w^2_i(f)}}}.


    If main_spectra is 2D and has dimensions N_1 x N_2, final spectrum has dimension N_2 (in contrast to `calc_Y_sigma_from_Yf_sigmaf`
    which combines across other dimension).

    Parameters
    =========
    main_spectra: np.array
        Array of arrays or FrequencySeries or OmegaSpectrum objects to be combined.
    weights_spectra: np.array
        Array of arrays or FrequencySeries or OmegaSpectrum objects to use as weights.

    Returns
    =======
    combined_weighted_spectrum: array_like
        Final spectrum obtained combining the original spectra with given weights.
    combined_weights_spectrum: array_like
        Variance associated to the final spectrum obtained combining the given weights.
    """
    if isinstance(main_spectra[0], OmegaSpectrum):
        _check_omegaspectra(main_spectra)

    if isinstance(weights_spectra[0], OmegaSpectrum):
        _check_omegaspectra(weights_spectra)

    w_spec = np.array(weights_spectra)
    m_spec = np.array(main_spectra)
    res_1 = 1 / np.nansum(1 / w_spec ** 2, axis=0)
    combined_weights_spectrum = np.sqrt(res_1)
    combined_weighted_spectrum = (
        np.nansum(m_spec / w_spec ** 2, axis=0) * res_1
    )
    if isinstance(main_spectra[0], OmegaSpectrum):
        combined_weighted_omegaspectrum = OmegaSpectrum(combined_weighted_spectrum, alpha=main_spectra[0].alpha, fref=main_spectra[0].fref, h0=main_spectra[0].h0, frequencies=main_spectra[0].frequencies, name='omega_spectrum')
        combined_weights_omegaspectrum = OmegaSpectrum(combined_weights_spectrum, alpha=main_spectra[0].alpha, fref=main_spectra[0].fref, h0=main_spectra[0].h0, frequencies=main_spectra[0].frequencies, name='sigma_spectrum')
        return combined_weighted_omegaspectrum, combined_weights_omegaspectrum
    else:
        return combined_weighted_spectrum, combined_weights_spectrum



from astropy.constants import c
from astropy.cosmology import Planck18 as cosmo

H0 = cosmo.H(0)
h0 = H0.value / 100.0
deprecated_H0 = 3.240779289444365023237687716352957261e-18  # LAL_H0FAC_SI = 100 km/s/Mpc in SI units
# H0=3.2407792903e-18 # stochastic.m value
speed_of_light = c.value  # m/s



# -*- coding: utf-8 -*-
# Copyright (C) California Institute of Technology 2023
#
# This file is part of pygwb.

import argparse
import glob
import logging
import os

import gwsumm
import gwsumm.plot
import gwsumm.tabs
from gwpy.time import from_gps

"""Create pygwb html
"""

def pygwb_html(outdir='./', config=None, segment_results=False):

    # =====================
    # get data
    
    ifo = 'Network'
    # =====================
    # now make the html page

    os.chdir('/'.join(outdir.split('/')[:-1]))
    output_dir = outdir.split('/')[-1]
    
    tabs = []
    base_tabs = []
    all_plots = []
    
    # now make main tab
    logging.info('Processing main results page...')
    home_index = os.path.join(output_dir,'index.html')

    plot_sub_dir = os.path.join(output_dir, 'output/combined_results')
    home_plots = glob.glob(plot_sub_dir+'/*.png')
    home_plots = sorted(home_plots)
    home_plots = ['/'.join(plot_loc.split('/')[-3:]) for plot_loc in home_plots]
    
    plot_tab = gwsumm.tabs.PlotTab('pygwb combined results',path=output_dir,index=home_index)
    plot_tab.set_layout((2,2))
    for plot_name in home_plots:
        plot = gwsumm.plot.core.SummaryPlot(href='./'+os.path.join(output_dir,plot_name))

        plot_tab.add_plot(plot)
    base_tabs.append(plot_tab)
    tabs.append(plot_tab)

    seg_tabs = []
    if segment_results:
        segment_folders = glob.glob(output_dir+'/output/*-*')
        for i, seg_folder in enumerate(segment_folders):
            page_name = seg_folder.split('/')[-1].split('-')[0]
            page_name = from_gps(int(page_name)).strftime('%Y-%m-%d %H:%M:%S')
            if i==0:
                tab0 = gwsumm.tabs.PlotTab(page_name,
                                           path=output_dir,parent='Segment results')
                tab = tab0
            else:
                tab = gwsumm.tabs.PlotTab(page_name,
                                          path=output_dir,parent=tab0.parent)

            seg_plots = glob.glob(seg_folder+'/*.png')
            seg_plots = sorted(seg_plots)
            seg_plots = ['/'.join(plot_loc.split('/')[-3:]) for plot_loc in seg_plots]
            for plot_name in seg_plots:
                plot = gwsumm.plot.core.SummaryPlot(href='./'+os.path.join(output_dir,plot_name))
                tab.add_plot(plot)
            tab.set_layout((2))
            seg_tabs.append(tab)
            tabs.append(tab)
    
    about_tab = gwsumm.tabs.AboutTab('About',path=output_dir)
    about_href = about_tab.href
    
    # Write out tabs
    logging.info('Writing html of tabs...')
    for tab in base_tabs:
        tab.write_html('',tabs=tabs,about=about_href)

    for tab in seg_tabs:
        tab.write_html('',tabs=tabs,about=about_href)
    
    if config:
        about_tab.write_html(tabs=tabs, config=config) # fix me

def main(args=None):
    logging.basicConfig()
    logger = logging.getLogger(__process_name__)
    logger.setLevel(logging.DEBUG)

    parser = argparse.ArgumentParser(description=__doc__,
                                     prog=__process_name__)
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='increase verbose output')
    parser.add_argument('-V', '--version', action='version',
                        version=__version__)
    parser.add_argument('-o', '--output-dir', type=os.path.abspath,
                        help='Directory for all output')
    parser.add_argument('-p', '--plot-dir', type=os.path.abspath,
                        help='Directory of plots to show')
    parser.add_argument('--plot-segment-results', action='store_true',
                        help='Create result pages for every segment')
    args = parser.parse_args(args=args)

    # call the above function
    pygwb_html(outdir=args.output_dir, segment_results=args.plot_segment_results)

# allow be be run on the command line
if __name__ == "__main__":
    main(args=None)




import numpy as np
from loguru import logger

from pygwb.postprocessing import calculate_point_estimate_sigma_spectra
from pygwb.util import calc_bias


def dsc_cut(
    naive_sigma: np.ndarray,
    slide_sigma: np.ndarray,
    dsc: float = 0.2,
    bf_ss: float = 1,
    bf_ns: float = 1,
):
    """
    Function that performs the delta sigma cut, a veto that marks certain GPS times as unusable if the estimations of
    the PSD in the naive (estimating sigma in bin J) and sliding (estimating sigma in bins J \pm 1) differ by more than
    a certain factor (default: dsc=0.2)

    Parameters
    ==========
    naive_sigma: array
        Naive sigma

    slide_sigma: array
        Sliding sigma

    dsc: float
        Threshold to perform the delta sigma cut

    bf_ss: float
        Sliding bias factor

    bf_ns: float
        Naive bias factor

    Returns
    =======
    dsigma >= dsc: bool
        True: the segment's delta sigma exceeds the threshold value, thus making its corresponding GPStime BAD
        False:  the segment's delta sigma is less than the threshold value, thus making its corresponding GPStime GOOD
    dsigma: np.array
        values of the difference between sliding sigma and naive sigma, i.e.: real value of the delta sigma cut per segment
    """

    dsigma = np.abs(slide_sigma * bf_ss - naive_sigma * bf_ns) / (slide_sigma * bf_ss)

    return dsigma >= dsc, dsigma


def run_dsc(
    dsc: float,
    segment_duration: int,
    psd1_naive: np.ndarray,
    psd2_naive: np.ndarray,
    psd1_slide: np.ndarray,
    psd2_slide: np.ndarray,
    alphas: np.ndarray,
    sample_rate: int,
    orf: np.array,
    fref: int,
    frequency_mask: np.array = True,
    window_fftgram_dict: dict = {"window_fftgram": "hann"},
    overlap_factor: float=0.5,
    N_average_segments_psd = 2,
    return_naive_and_averaged_sigmas: bool = False,
):

    """
    Function that runs the delta sigma cut

    Parameters
    ==========
    dsc: float
        The value of the delta sigma cut to use

    segment_duration: int
        Duration of each segment

    psd1_naive; psd2_naive: np.array
        an FFTgram of the PSD computed naively, as in in the particular bin J for detector #1 and #2

    psd1_slide, psd2_slide: np.array
        an FFTgram of the PSD computed by considering the noise in adjacent bins to the bin J, i.e. J-1, J+1 for
        detectors #1 and #2

    alphas: np.array
        the spectral indices to use; the code combines the BadGPStimes from each alpha

    sample_rate: int
        sampling rate (Hz)

    notch_list_path: str
        path to the notch list file

    window_fftgram_dict: dictionary, optional
        Dictionary with window characteristics. Default is `(window_fftgram_dict={"window_fftgram": "hann"}`

    orf: array
        the overlap reduction function as a function of frequency that quantifies the overlap of a detector baseline,
        which depends on the detector locations, relative orientations, etc.

    fref: int
        reference frequency (Hz)

    return_naive_and_averaged_sigmas: bool
        option to return naive and sliding sigmas

    Returns
    =======
    BadGPStimes: np.array
        an array of the GPS times to not be considered based on the chosen value of the delta sigma cut
    """

    logger.info("Running delta sigma cut")
    nalphas = len(alphas)
    times = np.array(psd1_naive.times)
    ntimes = len(times)
    df = psd1_naive.df.value
    dt = 1 / sample_rate  
    # Naive estimate
    bf_ns = calc_bias(
        segmentDuration=segment_duration,
        deltaF=df,
        deltaT=dt,
        N_avg_segs=1,
        window_fftgram_dict=window_fftgram_dict,
        overlap_factor=overlap_factor
    )
    # Sliding estimate
    bf_ss = calc_bias(
        segmentDuration=segment_duration,
        deltaF=df,
        deltaT=dt,
        N_avg_segs=N_average_segments_psd,
        window_fftgram_dict=window_fftgram_dict,
        overlap_factor=overlap_factor
    )
    freqs = np.array(psd1_naive.frequencies)
    overall_cut = np.zeros((ntimes, 1), dtype="bool")
    cuts = np.zeros((nalphas, ntimes), dtype="bool")
    naive_sigmas = np.zeros((nalphas, ntimes))
    slide_sigmas = np.zeros((nalphas, ntimes))
    dsigmas = np.zeros((nalphas, ntimes))

    for idx, alpha in enumerate(alphas):
        cut = np.zeros((ntimes, 1), dtype="bool")
        dsigma = np.zeros((ntimes, 1), dtype="float")
        for time in range(len(times)):
            psd1_naive_time = psd1_naive[time, :]
            psd1_slide_time = psd1_slide[time, :]
            psd2_naive_time = psd2_naive[time, :]
            psd2_slide_time = psd2_slide[time, :]

            naive_sigma_with_Hf = calculate_point_estimate_sigma_spectra(
                freqs=freqs,
                avg_psd_1=psd1_naive_time,
                avg_psd_2=psd2_naive_time,
                orf=orf,
                sample_rate=sample_rate,
                window_fftgram_dict=window_fftgram_dict,
                overlap_factor=overlap_factor,
                segment_duration=segment_duration,
                csd=None,
                fref=fref,
                alpha=alpha,
            )

            slide_sigma_with_Hf = calculate_point_estimate_sigma_spectra(
                freqs=freqs,
                avg_psd_1=psd1_slide_time,
                avg_psd_2=psd2_slide_time,
                orf=orf,
                sample_rate=sample_rate,
                window_fftgram_dict=window_fftgram_dict,
                overlap_factor=overlap_factor,
                segment_duration=segment_duration,
                csd=None,
                fref=fref,
                alpha=alpha,
            )

            naive_sensitivity_integrand_with_Hf = 1.0 / naive_sigma_with_Hf
            slide_sensitivity_integrand_with_Hf = 1.0 / slide_sigma_with_Hf

            naive_sigma_alpha = np.sqrt(
                1 / np.sum(naive_sensitivity_integrand_with_Hf[frequency_mask])
            )
            slide_sigma_alpha = np.sqrt(
                1 / np.sum(slide_sensitivity_integrand_with_Hf[frequency_mask])
            )
            naive_sigmas[idx, time] = naive_sigma_alpha
            slide_sigmas[idx, time] = slide_sigma_alpha

            cut[time], dsigma[time] = dsc_cut(
                naive_sigma=naive_sigma_alpha,
                slide_sigma=slide_sigma_alpha,
                dsc=dsc,
                bf_ss=bf_ss,
                bf_ns=bf_ns,
            )

        cuts[idx, :] = np.squeeze(cut)
        dsigmas[idx, :] = np.squeeze(dsigma)

    for time in range(len(times)):
        overall_cut[time] = any(cuts[:, time])

    BadGPStimes = times[np.squeeze(overall_cut)]

    dsigmas_dict = {}
    dsigmas_dict["alphas"] = alphas
    dsigmas_dict["times"] = times
    dsigmas_dict["values"] = dsigmas

    if return_naive_and_averaged_sigmas == True:
        dsigmas_dict["slide_sigmas"] = slide_sigmas * bf_ss
        dsigmas_dict["naive_sigmas"] = naive_sigmas * bf_ns

    return BadGPStimes, dsigmas_dict



import os
import warnings

import gwpy
import numpy as np
from loguru import logger

from .baseline import Baseline
from .notch import StochNotchList
from .omega_spectra import OmegaSpectrum
from .postprocessing import (
    calc_Y_sigma_from_Yf_sigmaf,
    combine_spectra_with_sigma_weights,
)
from .simulator import Simulator


class Network(object):
    """
    Network object for stochastic analyses with multiple functionalities:
        * data simulation
        * stochastic pre-processing
        * isotropic stochastic analysis
    """

    def __init__(
        self,
        name,
        interferometers,
        duration=None,
        frequencies=None,
        calibration_epsilon=0,
        notch_list_path="",
        coarse_grain_psd=False,
        coarse_grain_csd=True,
        overlap_factor_welch=0.5,
        overlap_factor=0.5,
        window_fftgram_dict={"window_fftgram": "hann"},
        window_fftgram_dict_welch={"window_fftgram": "hann"},
        N_average_segments_psd=2,
    ):
        """
        Instantiate a Network object.

        Parameters
        ----------
        name: str
            Name for the network, e.g H1H2
        interferometers: list of intereferometer objects
        duration: float, optional
            the duration in seconds of each data segment in the interferometers. None by default, in which case duration is inherited from the interferometers.
        frequencies: array_like, optional
            the frequency array for the Baseline and
            interferometers
        calibration_epsilon: float, optional
            calibration uncertainty for this baseline -- currently only supports a single notch list for all baselines
        notch_list_path: str, optional
            file path of the baseline notch list -- currently only supports a single notch list for all baselines
        coarse_grain_psd: bool
            Whether to apply coarse graining to obtain PSD spectra. Default is False.
        coarse_grain_csd: bool
            Whether to apply coarse graining to obtain CSD spectra. Default is True.
        overlap_factor_welch: float, optional
            Overlap factor to use when if using Welch's method to estimate spectra (NOT coarsegraining). For \"hann\" window use 0.5 overlap_factor and for \"boxcar"\ window use 0 overlap_factor. Default is 0.5 (50% overlap), which is optimal when using Welch's method with a \"hann\" window.
        overlap_factor: float, optional
            Factor by which to overlap the segments in the psd and csd estimation.
            Default is 1/2, if set to 0 no overlap is performed.
        window_fftgram_dict: dictionary, optional
            Dictionary containing name and parameters describing which window to use when producing fftgrams for csds (and psds if these are coarse-grained). Default is \"hann\".
        window_fftgram_dict_welch: dictionary, optional
            Dictionary containing name and parameters describing which window to use when producing fftgrams with Welch's method. Default is \"hann\".
        N_average_segments_psd: int, optional
            Number of segments used for PSD averaging (from both sides of the segment of interest)
            N_avg_segs should be even and >= 2
        """
        self.name = name
        self.interferometers = interferometers
        self.Nifo = len(interferometers)
        self.set_duration(duration)

        combo_tuples = []
        for j in range(1, len(interferometers)):
            for k in range(j):
                combo_tuples.append((k, j))

        baselines = []
        for i, j in combo_tuples:
            base_name = (
                f"{self.interferometers[i].name} - {self.interferometers[j].name}"
            )
            baselines.append(
                Baseline(
                    base_name,
                    self.interferometers[i],
                    self.interferometers[j],
                    duration=duration,
                    frequencies=frequencies,
                    calibration_epsilon=calibration_epsilon,
                    notch_list_path=notch_list_path,
                    coarse_grain_psd=coarse_grain_psd,
                    coarse_grain_csd=coarse_grain_csd,
                    overlap_factor_welch=overlap_factor_welch,
                    overlap_factor=overlap_factor,
                    window_fftgram_dict=window_fftgram_dict,
                    window_fftgram_dict_welch=window_fftgram_dict_welch,
                    N_average_segments_psd=N_average_segments_psd,
                )
            )

        self.baselines = baselines

    @classmethod
    def from_baselines(cls, name, baselines):
        """
        Initialise a network from a set of baselines. Takes care to unpack the interferometers from each baselines and sets them in the Network.
        """
        if not all(baselines[0].duration == base.duration for base in baselines[1:]):
            raise AssertionError(
                "All baselines used to initialise must have same duration set."
            )
        network = cls(
            name,
            interferometers=[],
            duration=baselines[0].duration,
        )
        network.baselines = baselines
        ifo_visited = set()
        interferometers = []
        for base in baselines:
            if base.interferometer_1.name not in ifo_visited:
                ifo_visited.add(base.interferometer_1.name)
                interferometers.append(base.interferometer_1)
            if base.interferometer_2.name not in ifo_visited:
                ifo_visited.add(base.interferometer_2.name)
                interferometers.append(base.interferometer_2)
        network.interferometers = interferometers
        network.Nifo = len(interferometers)
        return network

    def set_duration(self, duration):
        """Sets the duration for the Network and Interferometers

        Note: the cross-checks that durations match in all the interferometers are done by each Baseline.

        Parameters
        ==========
        duration: float, optional
            The duration to set for the Network and interferometers
        """
        if duration is not None:
            self.duration = duration
            for ifo in self.interferometers:
                ifo.duration = duration
            return
        try:
            duration = next(
                ifo.duration for ifo in self.interferometers if ifo.duration is not None
            )
        except StopIteration:
            warnings.warn(
                "The Network duration is not set, "
                "and the interferometer durations are all None."
            )
            self.duration = None
            return

        check_dur = all(ifo.duration == duration for ifo in self.interferometers)
        if not check_dur:
            raise AssertionError(
                "The interferometer durations don't match! "
                "The Network can't handle this. "
                "Make sure that the interferometer durations are the same."
            )
            # for ifo in self.interferometers:
            #    ifo.duration = duration
        self.duration = duration

    def set_frequency_mask(self, notch_list_path="", flow=20, fhigh=1726):
        """
        Set frequency mask to frequencies attribute.

        Parameters
        ==========
        notch_list_path: str
            Path to notch list to apply to frequency array.
        """
        mask = (self.frequencies >= flow) & (self.frequencies <= fhigh)
        if notch_list_path:
            notch_list = StochNotchList.load_from_file(notch_list_path)
            notch_mask = notch_list.get_notch_mask(self.frequencies)
            mask = np.logical_and(mask, notch_mask)
        self.frequency_mask = mask

    def set_interferometer_data_from_simulator(
        self,
        N_segments,
        GWB_intensity=None,
        CBC_dict=None,
        sampling_frequency=None,
        start_time=None,
        inject_into_data_flag=False,
    ):
        """
        Fill interferometers with data from simulation. Data can already be present in the intereferometers of the network,
        in which case the simulated data will be injected on top of the data already present.

        Parameters
        ==========

        N_segments: int
            Number of segments to simulate
        GWB_intensity: gwpy.frequencyseries.FrequencySeries
            A gwpy.frequencyseries.FrequencySeries containing the desired strain power spectrum
        CBC_dict: dict
            Dictionary containing the parameters of CBC injections.
        sampling_frequency: float
            Sampling frequency at which the data needs to be simulated. If not specified (None), will check for interferometer's
            sampling frequency.
        start_time: float
            Start time of the simulated data. If not passed (None), will check for interferometer's timeseries start time.
            If not specified either, start time will default to zero.
        inject_into_data_flag: boolean
            Flag that specifies whether or not the simulated data needs to be injected into data, i.e. if there is already
            data present in the interferometers of the network. If so, only data will be simulated and no extra noise will
            be added on top of the simulated data.
        """
        no_noise = inject_into_data_flag
        if start_time == None:
            try:
                start_time = self.interferometers[0].timeseries.times.value[0]
            except:
                start_time = 0.0
                warnings.warn(
                    "User did not specify a start time. Setting start time to zero."
                )

        if sampling_frequency == None:
            try:
                sampling_frequency = self.interferometers[0].sampling_frequency
            except:
                raise ValueError(
                    "Sampling frequency was not set. Please set sampling frequency in interferometers or pass directly to function."
                )

        data_simulator = Simulator(
            self.interferometers,
            N_segments,
            duration=self.duration,
            intensity_GW=GWB_intensity,
            injection_dict=CBC_dict,
            start_time=start_time,
            sampling_frequency=sampling_frequency,
            no_noise=no_noise,
        )
        data = data_simulator.get_data_for_interferometers()

        if inject_into_data_flag:
            for ifo in self.interferometers:
                ifo.set_strain_data_from_gwpy_timeseries(
                    ifo.strain_data.to_gwpy_timeseries().inject(data[ifo.name])
                )
        else:
            for ifo in self.interferometers:
                ifo.set_strain_data_from_gwpy_timeseries(data[ifo.name])

    def save_interferometer_data_to_file(self, save_dir="./", file_format="hdf5"):
        """
        Save interferometer strain data to a file. This method relies on the gwpy  TimeSeries.write method. Typically used when simulating a signal for a whole network of interferometers.
        Note: this will save a single frame file with a set of interferometer data; each strain channel is labelled by its interferometer.

        Parameters
        ==========
        save dir: str, optional
            The path of the output folder. Defaults to the local folder.
        file format: str, optional
            The format of the output file. Defaults to hdf5 file. Acceptable formats are standard gwpy TimeSeries.write formats.
        """
        file_name = f"{self.name}_STRAIN-{int(self.interferometers[0].strain_data.start_time)}-{int(self.interferometers[0].strain_data.duration)}.{file_format}"
        file_path = os.path.join(save_dir, file_name)
        data_dict = gwpy.timeseries.TimeSeriesDict()
        for ifo in self.interferometers:
            channel = f"STRAIN_{ifo.name}"
            data_dict[channel] = ifo.strain_data.to_gwpy_timeseries()
        data_dict.write(file_path, format=file_format)

    def combine_point_estimate_sigma_spectra(self):
        """
        Combines the point estimate and sigma spectra from different baselines in the Network and stores them as attributes.
        """
        try:
            point_estimate_spectra = [
                base.point_estimate_spectrum for base in self.baselines
            ]
            sigma_spectra = [base.sigma_spectrum for base in self.baselines]
        except AttributeError:
            raise AttributeError("The Baselines of the Network have not been set!")

        alphas = np.array([spec.alpha for spec in point_estimate_spectra])
        frefs = np.array([spec.fref for spec in point_estimate_spectra])
        h0s = np.array([spec.h0 for spec in point_estimate_spectra])
        dfs = np.array([spec.df.value for spec in point_estimate_spectra])
        f0s = np.array([spec.f0.value for spec in point_estimate_spectra])

        dict_attributes = {
            "alpha": [alphas, "spectral indices"],
            "fref": [frefs, "reference frequencies"],
            "h0": [h0s, "cosmology h0"],
            "df": [dfs, "sampling frequency"],
            "f0": [f0s, "begin frequency"],
        }

        for key in dict_attributes:
            if not np.all(dict_attributes[key][0] == dict_attributes[key][0][0]):
                raise ValueError(
                    f"The {dict_attributes[key][1]} of the spectra in each Baseline don't match! Spectra may not be combined."
                )

        pt_est_spec, sig_spec = combine_spectra_with_sigma_weights(
            np.array(point_estimate_spectra), np.array(sigma_spectra)
        )
        self.point_estimate_spectrum = OmegaSpectrum(
            pt_est_spec,
            alpha=alphas[0],
            fref=frefs[0],
            h0=h0s[0],
            name="Y_spectrum_network",
            frequencies=self.baselines[0].frequencies,
        )
        self.sigma_spectrum = OmegaSpectrum(
            sig_spec,
            alpha=alphas[0],
            fref=frefs[0],
            h0=h0s[0],
            name="sigma_spectrum_network",
            frequencies=self.baselines[0].frequencies,
        )

    def set_point_estimate_sigma(
        self,
        notch_list_path="",
        flow=20,
        fhigh=1726,
    ):
        """
        Set point estimate sigma based the combined spectra from each Baseline. This is estimate of omega_gw in each frequency bin.

        Parameters
        ==========
        notch_list_path: str, optional
            Path to the notch list to use in the spectrum; if the notch_list isn't set in the baseline,
            user can pass it directly here. If it is not set and if none is passed no notches will be applied.
        flow: float, optional
            Low frequency. Default is 20 Hz.
        fhigh: float, optional
            High frequency. Default is 1726 Hz.
        """
        if not hasattr(self, "point_estimate_spectrum"):
            logger.info(
                "Point estimate and sigma spectra have not been set before. Setting it now..."
            )
            self.combine_point_estimate_sigma_spectra()

        self.frequencies = self.point_estimate_spectrum.frequencies.value
        self.set_frequency_mask(notch_list_path, flow=flow, fhigh=fhigh)

        Y, sigma = calc_Y_sigma_from_Yf_sigmaf(
            self.point_estimate_spectrum,
            self.sigma_spectrum,
            frequency_mask=self.frequency_mask,
        )

        self.point_estimate = Y
        self.sigma = sigma


# TODO: add options in the network to apply dsc, frequency notching at the network level.



# -*- coding: utf-8 -*-
# Copyright (C) California Institute of Technology 2023
#
# This file is part of pygwb.

"""Core workflow functions"""

import configparser
import logging
import os
import re
import shutil
from collections.abc import Iterable
from getpass import getuser

import numpy as np
from gwpy.io.cache import cache_segments
from gwpy.segments import DataQualityDict, DataQualityFlag, Segment, SegmentList
from gwsumm.data.timeseries import find_best_frames
from pycondor import Dagman as pyDagman
from pycondor import Job as pyJob
from pycondor.job import JobArg

ACCOUNTING_GROUP_USER = os.getenv(
    '_CONDOR_ACCOUNTING_USER',
    getuser(),
)

def makedir(dir_path):
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)

class Job(pyJob):
    """
    Wrapper class of pycondor.Job
    """

    def __init__(self, name, executable, output_arg=None, 
                 output_file=None, accounting_group=None, 
                 extra_lines=[], arguments=None,retry=None,
                 final_result=False, input_file=None,
                 request_disk='2048MB', request_memory='512MB',
                 required_job=False,
                 **kwargs):

        exec_path = shutil.which(executable)
        if exec_path is None:
            raise TypeError('execuatble must be installed in environment'+
                       ' or given as an absolute path')

        if accounting_group is not None:
            condorcmds = [
                f"accounting_group = {accounting_group}",
                f"accounting_group_user = {ACCOUNTING_GROUP_USER}",
                f"getenv = True",
                ]
            condorcmds.extend(extra_lines)
        else:
            raise TypeError('accounting group must be supplied')

        output_args = []
        self.output_file = None
        # attrs for cache system
        self.output_exits = False
        self.required_job = False
            
        # output args
        if output_file is not None:
            if output_arg:
                if isinstance(output_arg, str):
                    output_args.append(output_arg)
                else:
                    raise TypeError('output_arg must be a string')
            if isinstance(output_file, str):
                self.output_file = [output_file]
                if output_arg:
                    output_args.append(output_file)
            elif isinstance(output_file, Iterable):
                self.output_file = output_file
                if output_arg:
                    for arg in output_file:
                        output_args.append(arg) 
            else:
                raise TypeError('output_files must be a string or an iterable')

        #input_args
        self.input_file = None
        if input_file is not None:
            if isinstance(input_file, str):
                self.input_file = [input_file]
            elif isinstance(input_file, Iterable):
                self.input_file = input_file

        # final result? 
        self.final_result = final_result
                
        arguments = ' '.join(list([arguments])+list(output_args))

        # surround dicts with ""
        # THIS HAS BEEN SOMEHOW DEPRECATED: CONDOR ADDS THEM AUTOMATICALLY
        #arguments = arguments.replace("{", "'{").replace("}", "}'")

        # convert memory into integer
        try:
            request_memory = int(request_memory)
        except:
            # check if it is MB, GB, or KB
            if 'M' in request_memory:
                request_memory = int(re.findall('\d+', request_memory)[0])
            elif 'G' in request_memory:
                request_memory = int(re.findall('\d+', request_memory)[0]) * 1024
            else:
                raise TypeError('request_memory must be an integer or in format of e.g. 1024MB or 1GB')

        # release if requested memory is too small
        request_memory = f"ifthenelse(isUndefined(MemoryUsage),{request_memory},3*MemoryUsage)"
        condorcmds.extend(["periodic_release = (HoldReasonCode == 26) && (JobStatus == 5)"])

        if retry == None:
            retry = 3
            
        super(Job,self).__init__(name, exec_path, arguments=arguments,
                                 extra_lines=condorcmds, retry=retry, 
                                 request_disk=request_disk, 
                                 request_memory=request_memory, 
                                 **kwargs)

    def check_required(self):
        if not self.output_exits:
            self.required_job = True
            for p in self.parents:
                if not p.required_job:
                    p.check_required()

    def replace_input(self, cache):
        args = self.args[0].arg.split(' ')
        input_base = []
        if self.input_file:
            input_base = [os.path.basename(i_file) for i_file in self.input_file]
        for file_pair in cache:
            if file_pair[0] in input_base:
                # replace this in arguments
                for i, arg in enumerate(args):
                    if file_pair[0] in arg:
                        args[i] = file_pair[1]
        self.args = [JobArg(arg=' '.join(args),
                            name=self.args[0].name,
                            retry=self.args[0].retry)]

    def replace_input_children(self, cache):
        for c in self.children:
            c.replace_input(cache)

class Dagman(pyDagman):
    """
    Wrapper class of pycondor.Dagman
    """
    def __init__(self, name, basedir=None, **kwargs):
        self.base_dir = os.path.abspath(basedir)
        self.submit_dir = os.path.join(self.base_dir, 'condor')
        self.error_dir = os.path.join(self.base_dir, 'condor')
        self.log_dir = os.path.join(self.base_dir, 'condor')
        self.output_dir = os.path.join(self.base_dir, 'condor')
        self.result_dir = os.path.join(self.base_dir, 'output')
        # used for caching
        self.input_cache_files = []
        super(Dagman,self).__init__(name, submit=self.submit_dir, **kwargs)

    def make_dir(self):
       makedir(self.base_dir)
       makedir(self.submit_dir)
       makedir(self.error_dir)
       makedir(self.log_dir)
       makedir(self.output_dir)
       makedir(self.result_dir)

    def remove_node(self, job):
        self.nodes.remove(job)

    def remove_parent(self, child_job, parent_job):
        child_job.parents.remove(parent_job)
    
    def in_cache(self, output_files, cache_list):
        if output_files:
            return all(os.path.basename(out) in cache_list.T[0]
                       for out in output_files)
        else:
            return False
    
    def add_to_dagman_inputs(self, file_names, cache):
        for file_name in file_names:
            f_name = os.path.basename(file_name)
            f_num = np.argwhere(cache.T[0] == f_name)
            self.input_cache_files.append(cache[f_num].flatten())
    
    def cache_filter(self, cache):
        # first figure out what jobs already have their outputs
        premade_jobs = []
        # try starting with "final result" jobs
        for job in self.nodes:
            #if all(out in cache for out in job.output_file):
            if self.in_cache(job.output_file, cache):
                #premade_jobs.append(job)
                job.output_exits = True
                job.replace_input_children(cache)
                self.add_to_dagman_inputs(job.output_file, cache)
        # then sort through nodes
        for job in self.nodes:
            if job.final_result:
                job.check_required()
    
    def remove_completed_jobs(self):
        node_list = self.nodes.copy()
        for j in node_list:
            if not j.required_job:
                # first remove children
                for c in j.children:
                    self.remove_parent(c, j)
                # then remove the node
                self.remove_node(j)
    
    def filter_and_remove_from_cache(self, cache):
        self.cache_filter(cache)
        self.remove_completed_jobs()
    
    def load_cache(self, cache_file):
        cache = np.loadtxt(cache_file, dtype=str)
        return cache[list(map(os.path.isfile,cache.T[1]))]
    
    def save_output_loc(self, output_loc='pygwb_cache.txt'):
        file_names = []
        file_paths = []
        for job in self.nodes:
            if job.output_file:
                file_names += [os.path.basename(out) for out in job.output_file]
                file_paths += job.output_file
        cache_output = np.array([file_names, file_paths]).T
        if len(self.input_cache_files):
            cache_output = np.concatenate((cache_output, self.input_cache_files))
        np.savetxt(output_loc,
                   cache_output,
                   delimiter=" ", fmt="%s")

    def run_jobs_serially(self):
        import subprocess

        # run everything locally and serially
        remaining_nodes = self.nodes.copy()
        node_order = []
        node_step = 0

        # first add jobs with no parents
        for node in remaining_nodes:
            if not len(node.parents):
                node_order.append([node.name, node.executable, node.args[0].arg])
                remaining_nodes.remove(node)

        # now check the other nodes
        while len(remaining_nodes):
            for node in remaining_nodes:
                if all([n not in remaining_nodes for n in node.parents]):
                    node_order.append([node.name, node.executable, node.args[0].arg])
                    remaining_nodes.remove(node)
            node_step += 1
            #double check that we aren't in an inf loop
            if node_step > 100:
                raise RuntimeError('Too many steps when trying '
                                   'to run serially!')

        logging.info('Beginning to run all jobs serially...')
        node_num = 1
        node_len = len(node_order)
        for node in node_order:
            path_and_args = node[1] + ' ' + node[2]
            logging.info(f'Starting node #{node_num}/{node_len}: {node[0]}')
            logging.info(f'The full path and arguments are: {path_and_args}')
            subprocess.run(path_and_args, check=True, shell=True)
            node_num += 1
        logging.info('Workflow completed!')

def _collect_job_arguments(config, job_type):
    config_sec = config[job_type]
    args_list = list()
    for key in config_sec.keys():
        args_list.append('--' + str(key))
        val = str(config_sec[key])
        if val[0] == '$':
            val_source = val.replace('${', '').replace('}', '')
            val_source = val_source.split(':')
            val = config[val_source[0]][val_source[1]]
        args_list.append(val)
    return args_list

class Workflow():
    def __init__(self, name, config_file, basedir='./workflow', 
                 config_overrides=[]):
        self.t0 = None
        self.tf = None

        self.config = configparser.ConfigParser()
        self.config.read(config_file)
        self.config_override(config_overrides)

        self.cache_file_path = None
        if self.config.has_option('general', 'cache_file'):
            self.cache_file_path = os.path.abspath(self.config['general']['cache_file'])

        self.setup_dagman(name, basedir=basedir) 

    def config_override(self, overrides):
        for ov in overrides:
            split_ov = ov.split(':')
            try:
                ov_sec, ov_key, ov_val = \
                    split_ov[0], split_ov[1], ':'.join(split_ov[2:])
            except:
                raise ValueError(f"Overrides must be in the form 'section:key:value', you used '{ov}'")
            if ov_sec not in self.config.sections():
                self.config[ov_sec] = {}
            self.config[ov_sec][ov_key] = ov_val

    def config_write(self, output_loc):
        with open(output_loc, 'w') as configfile:
            self.config.write(configfile)

    def setup_segments(self):
        self.t0 = int(self.config['general']['t0'])
        self.tf = int(self.config['general']['tf'])
        dur = int(self.tf-self.t0)

        if self.config.has_option('general', 'min_job_dur'):
            self.min_dur = int(self.config['general']['min_job_dur'])
        else:
            self.min_dur = 0
        if self.config.has_option('general', 'max_job_dur'):
            self.max_dur = int(self.config['general']['max_job_dur'])
        else:
            self.max_dur = 1000000

        ifos = self.config['general']['ifos'].split(' ')
        strfr = self.config["pygwb_pipe"]["frametype"].replace(":", ",").split(",")
        strain_frametypes = {strfr[2*i]: strfr[(2*i+1)] for i in range(len(strfr)//2)}

        seglist = SegmentList([[self.t0, self.tf]])
        for ifo in ifos:
            frametype = strain_frametypes[ifo]
            cached_frames, _ = find_best_frames(ifo, frametype, start=self.t0, end=self.tf)
            seglist &= cache_segments(cached_frames)

        logging.info('Downloading science flags...')
        sci_flag = DataQualityDict.query_dqsegdb(
            [i+':'+self.config['data_quality']['science_segment'] for i in ifos],
            self.t0, self.tf
            ).intersection().active

        if self.config.has_option('data_quality', 'veto_definer'):
            logging.info('Downloading vetoes...')
            vetoes = DataQualityDict.from_veto_definer_file(
                config['data_quality']['veto_definer']
                )
            cat1_vetoes = DataQualityDict({v: vetoes[v] for v in vetoes if (vetoes[v].category ==1) and (v[:2] in ifos)})
            cat1_vetoes.populate()

            cat1_vetoes = cat1_vetoes.intersection().active

            seglist_flags = sci_flag - cat1_vetoes
        else:
            seglist_flags = sci_flag

        start_times = np.arange(self.t0, self.tf, self.max_dur)
        if len(start_times) > 1:
            start_times = start_times[:-1]

        block_segs = SegmentList(
            [
                Segment(t, t + self.max_dur)
                for t in start_times
            ]
        )
        block_segs &= seglist

        seglist_pruned = []
        for seg in block_segs:
            if seglist_flags.intersects_segment(seg):
                new_seg_list = (SegmentList([seg]) & seglist_flags)
                for new_seg in new_seg_list:
                    if abs(new_seg) > self.min_dur:
                        seglist_pruned.append([int(new_seg[0]), int(new_seg[1])])

        return seglist_pruned


    def setup_dagman(self, name, basedir=None):
        logging.info('Writing DAG...')
        self.dagman = Dagman(name=name,
                        basedir=basedir)
        self.dagman.make_dir()

        self.config_path = os.path.join(self.dagman.base_dir, 'config.ini')
        self.config_write(self.config_path)

class IsotropicWorkflow(Workflow):
    def __init__(self, name, config_file, basedir='./workflow',
                 config_overrides=[]):
        super(IsotropicWorkflow, self).__init__(name, 
                                                config_file, basedir=basedir,
                                                config_overrides=config_overrides)
        self.fref = int(self.config['pygwb_pipe']['fref'])
        self.alpha =  float(self.config['pygwb_pipe']['alpha'])

    def create_pygwb_pipe_job(self, t0=None, tf=None, parents=[], output_path=None):
    #pygwb_pipe --param_file parameters.ini --output_path output/ --t0 ${START} --tf ${END}
        dur = int(tf-t0)
        file_tag = f'{int(t0)}-{dur}'
        name = f'pygwb_pipe_{int(t0)}_{dur}'
        args = _collect_job_arguments(self.config, 'pygwb_pipe')
        args = args + [
            '--output_path', output_path,
            '--t0', str(t0),
            '--tf', str(tf),
            '--file_tag', file_tag
            ]
        output_file = [os.path.join(output_path, f'parameters_{file_tag}_final.ini'),
                       os.path.join(output_path, f'psds_csds_{file_tag}.npz'),
                       os.path.join(output_path, f'point_estimate_sigma_{file_tag}.npz')]
        job = Job(name=name,
                  executable=self.config['executables']['pygwb_pipe'],
                  accounting_group = self.config['general']['accounting_group'],
                  submit=self.dagman.submit_dir,
                  error=self.dagman.error_dir,
                  output=self.dagman.output_dir,
                  log=self.dagman.log_dir,
                  arguments=' '.join(args),
                  request_disk='128MB',
                  request_memory='2048MB',
                  output_file=output_file,
                  dag=self.dagman)
        for parent in parents:
            job.add_parent(parent)
        return job, output_file
    
    # FIX THE JOB LIST
    def create_pygwb_combine_job(self, t0=None, tf=None, parents=[], input_file=None, output_path=None,
        data_path=None, coherence_path=None, param_file=None,
        alpha=0., fref=30, h0=0.7):
    #pygwb_combine --data_path output/ --param_file output/parameters_final.ini --alpha 0 --fref 30 --h0 0.7 --out_path ./output/
        dur = int(tf-t0)
        file_tag = f'{int(t0)}-{dur}'
        name = f'pygwb_combine_{int(t0)}_{dur}'
        args = _collect_job_arguments(self.config, 'pygwb_combine')
        if isinstance(data_path, str):
            data_path = [data_path]
        if isinstance(coherence_path, str):
            coherence_path = [coherence_path]
    
        input_file = data_path + coherence_path + [param_file]
    
        data_path = ' '.join(data_path)
        coherence_path = ' '.join(coherence_path)
    
        args = args + [
            '--param_file', param_file,
            '--out_path', output_path,
            '--data_path', data_path,
            '--coherence_path', coherence_path,
            '--file_tag', file_tag
            ]
        output_file = [os.path.join(output_path, f'point_estimate_sigma_spectra_alpha_{alpha:.1f}_fref_{int(fref)}_{file_tag}.npz'),
                       os.path.join(output_path, f'coherence_spectrum_{file_tag}.npz'),
                       os.path.join(output_path, f'delta_sigma_cut_{file_tag}.npz')]
        job = Job(name=name,
                  executable=self.config['executables']['pygwb_combine'],
                  accounting_group = self.config['general']['accounting_group'],
                  submit=self.dagman.submit_dir,
                  error=self.dagman.error_dir,
                  output=self.dagman.output_dir,
                  log=self.dagman.log_dir,
                  arguments=' '.join(args),
                  request_disk='128MB',
                  request_memory='1024MB',
                  final_result=True,
                  input_file=input_file,
                  output_file=output_file,
                  dag=self.dagman)
        for parent in parents:
            job.add_parent(parent)
        return job, output_file
    
    def create_pygwb_stats_job(self, t0=None, tf=None, parents=[], output_path=None,
        csd_file=None, dsc_file=None, coh_file=None, param_file=None):
        dur = int(tf-t0)
        name = f'pygwb_stats_{int(t0)}_{dur}'
        args = _collect_job_arguments(self.config, 'pygwb_stats')
        args = args + [
            '-pf', param_file,
            '-pd', output_path,
            '-c', csd_file,
            '-dsc', dsc_file,
            '-fcoh', coh_file
            ]
        input_file = [param_file, csd_file, dsc_file, coh_file]
        # FIXME - figure out all the plots this job makes
        output_file = None
        job = Job(name=name,
                  executable=self.config['executables']['pygwb_stats'],
                  accounting_group = self.config['general']['accounting_group'],
                  submit=self.dagman.submit_dir,
                  error=self.dagman.error_dir,
                  output=self.dagman.output_dir,
                  log=self.dagman.log_dir,
                  arguments=' '.join(args),
                  request_disk='128MB',
                  request_memory='512MB',
                  final_result=True,
                  input_file=input_file,
                  output_file=output_file,
                  required_job=True,
                  dag=self.dagman)
        for parent in parents:
            job.add_parent(parent)
        return job
    
    def create_pygwb_html_job(self, t0=None, tf=None, segment_results=False, parents=[], output_path=None, plot_path=None):
        name = f'pygwb_html'
        args = _collect_job_arguments(self.config, 'pygwb_html')
        args = args + [
            '-p', plot_path,
            '-o', output_path,
            ]
        if segment_results:
            args = args + ['--plot-segment-results']
        job = Job(name=name,
                  executable=self.config['executables']['pygwb_html'],
                  accounting_group = self.config['general']['accounting_group'],
                  submit=self.dagman.submit_dir,
                  error=self.dagman.error_dir,
                  output=self.dagman.output_dir,
                  log=self.dagman.log_dir,
                  arguments=' '.join(args),
                  request_disk='128MB',
                  request_memory='1024MB',
                  final_result=True,
                  required_job=True,
                  dag=self.dagman)
        for parent in parents:
            job.add_parent(parent)
        return job
    



import pickle
import warnings

import numpy as np
from astropy.io import registry as io_registry
from gwpy.frequencyseries import FrequencySeries
from gwpy.spectrogram import Spectrogram
from gwpy.types.io.hdf5 import register_hdf5_array_io

from pygwb.constants import h0


class OmegaSpectrogram(Spectrogram):

    """Subclass of gwpy's Spectrogram class."""

    _metadata_slots = Spectrogram._metadata_slots + ("alpha", "fref", "h0")
    # _print_slots = FrequencySeries._print_slots + ["alpha", "fref", "h0"]

    def __new__(cls, data, **kwargs):
        kwargs.pop("alpha", None)
        kwargs.pop("fref", None)
        kwargs.pop("h0", None)
        return super(OmegaSpectrogram, cls).__new__(cls, data, **kwargs)

    def __init__(self, data, alpha=None, fref=None, h0=h0, **kwargs):
        if not isinstance(alpha, (float, int, np.number)):
            raise ValueError("Spectral index alpha must be a valid number.")
        if not isinstance(fref, (float, int, np.number)):
            raise ValueError("Reference frequency fref must be a valid number.")
        self._alpha = alpha
        self._fref = fref
        self._h0 = h0

    @property
    def alpha(self):
        """Spectral index alpha."""
        return self._alpha

    @alpha.setter
    def alpha(self, alpha):
        raise AttributeError(
            "alpha is protected! To change alpha, use the reweight method."
        )

    @property
    def fref(self):
        """Reference frequency."""
        return self._fref

    @fref.setter
    def fref(self, fref):
        raise AttributeError(
            "fref is protected! To change fref, use the reweight method."
        )

    @property
    def h0(self):
        """Hubble parameter h0. Default is pygwb.constants.h0 = 0.693199."""
        return self._h0

    @h0.setter
    def h0(self, h0):
        raise AttributeError("h0 is protected! To change h0, use the reset_h0 method.")

    @classmethod
    def read(cls, source, *args, **kwargs):
        """Read data into a Spectrogram. Same usage as read method of gwpy.Spectrogram.

        Parameters
        ==========
        source: str
            Source file path.
        """
        return io_registry.read(cls, source, *args, **kwargs)

    def write(self, target, *args, **kwargs):
        """Write this Spectrogram to a file. Same usage as write method of gwpy.Spectrogram.

        Parameters
        ==========
        target: str
            Target file path.
        """
        return io_registry.write(self, target, *args, **kwargs)

    @classmethod
    def load_from_pickle(cls, filename):
        """
        Load spectrogram object from pickle file.

        Parameters
        ==========
        filename: str
            Filename (inclusive of path) to load the pickled spectrogram from.
        """
        with open(filename, "rb") as f:
            return pickle.load(f)

    def save_to_pickle(self, filename):
        """
        Save spectrogram object to pickle file.

        Parameters
        ==========
        filename: str
            Filename (inclusive of path) to save the pickled spectrogram to.
        """
        with open(filename, "wb") as f:
            pickle.dump(self, f)

    def reweight(self, *, new_alpha=None, new_fref=None):
        """
        Reweight the spectrogram by a new spectral index alpha, and/or refer to a new reference freuency.

        Parameters:
        ==========

        new_alpha: float
            New spectral index.
        new_fref: float
            New reference frequency.
        """
        if new_alpha is None:
            new_alpha = self.alpha
        if new_fref is None:
            new_fref = self.fref
        for spectrum in self.value:
            new_spectrum = reweight_spectral_object(
                spectrum,
                self.frequencies,
                new_alpha,
                new_fref,
                old_alpha=self.alpha,
                old_fref=self.fref,
            )
            spectrum[:] = new_spectrum
        self._alpha = new_alpha
        self._fref = new_fref

    def reset_h0(self, new_h0):
        """
        Reset the hubble parameter h0. Expected values range between 0.5 and 1.

        Parameters:
        ==========

        new_h0: float
            New h0 to set the spectrum at.
        """
        if (new_h0 < 0.5) or (new_h0 > 1.0):
            warnings.warn(
                f"h0 should be between 0.5 and 1. The selected value of {new_h0} does not fall within this range."
            )
        new_spectrum = self.value * (self.h0 / new_h0) ** 2
        self.value[:] = new_spectrum
        self._h0 = new_h0


class OmegaSpectrum(FrequencySeries):

    """Subclass of gwpy's FrequencySeries class."""

    _metadata_slots = FrequencySeries._metadata_slots + ("alpha", "fref", "h0")
    _print_slots = FrequencySeries._print_slots + ["alpha", "fref", "h0"]

    def __new__(cls, data, **kwargs):
        kwargs.pop("alpha", None)
        kwargs.pop("fref", None)
        kwargs.pop("h0", None)
        return super(OmegaSpectrum, cls).__new__(cls, data, **kwargs)

    def __init__(self, data, alpha=None, fref=None, h0=h0, **kwargs):
        if not isinstance(alpha, (float, int, np.number)):
            raise ValueError("Spectral index alpha must be a valid number.")
        if not isinstance(fref, (float, int, np.number)):
            raise ValueError("Reference frequency fref must be a valid number.")
        self._alpha = alpha
        self._fref = fref
        self._h0 = h0

    @property
    def alpha(self):
        """Spectral index alpha."""
        return self._alpha

    @alpha.setter
    def alpha(self, alpha):
        raise AttributeError(
            "alpha is protected! To change alpha, use the reweight method."
        )

    @property
    def fref(self):
        """Reference frequency."""
        return self._fref

    @fref.setter
    def fref(self, fref):
        raise AttributeError(
            "fref is protected! To change fref, use the reweight method."
        )

    @property
    def h0(self):
        """Hubble parameter h0. Default is pygwb.constants.h0 = 0.693199."""
        return self._h0

    @h0.setter
    def h0(self, h0):
        raise AttributeError("h0 is protected! To change h0, use the reset_h0 method.")

    @classmethod
    def read(cls, source, *args, **kwargs):
        """Read data into a Spectrum. Same usage as read method of gwpy.FrequencySeries.

        Parameters
        ==========
        source: str
            Source file path.
        """
        return io_registry.read(cls, source, *args, **kwargs)

    def write(self, target, *args, **kwargs):
        """Write this Spectrum to a file. Same usage as write method of gwpy.FrequencySeries.


        Parameters
        ==========
        target: str
            Target file path.
        """
        return io_registry.write(self, target, *args, **kwargs)

    @classmethod
    def load_from_pickle(cls, filename):
        """
        Load spectrum object from pickle file.

        Parameters
        ==========
        filename: str
            Filename (inclusive of path) to load the pickled spectrum from.
        """
        with open(filename, "rb") as f:
            return pickle.load(f)

    def save_to_pickle(self, filename):
        """
        Save spectrum object to pickle file.

        Parameters
        ==========
        filename: str
            Filename (inclusive of path) to save the pickled spectrum to.
        """
        with open(filename, "wb") as f:
            pickle.dump(self, f)

    def reweight(self, *, new_alpha=None, new_fref=None):
        """
        Reweight the spectrum by a new spectral index alpha, and/or refer to a new reference freuency.

        Parameters:
        ==========

        new_alpha: float
            New spectral index.
        new_fref: float
            New reference frequency.
        """
        if new_alpha is None:
            new_alpha = self.alpha
        if new_fref is None:
            new_fref = self.fref
        new_spectrum = reweight_spectral_object(
            self.value,
            self.frequencies,
            new_alpha,
            new_fref,
            old_alpha=self.alpha,
            old_fref=self.fref,
        )
        self.value[:] = new_spectrum
        self._alpha = new_alpha
        self._fref = new_fref

    def reset_h0(self, new_h0):
        """
        Reset the hubble parameter h0. Expected values range between 0.5 and 1.

        Parameters:
        ==========

        new_h0: float
            New h0 to set the spectrum at.
        """
        if (new_h0 < 0.5) or (new_h0 > 1.0):
            raise ValueError("h0 must be between 0.5 and 1.")
        new_spectrum = self.value * (self.h0 / new_h0) ** 2
        self.value[:] = new_spectrum
        self._h0 = new_h0


def reweight_spectral_object(
    spec, freqs, new_alpha, new_fref, old_alpha=0.0, old_fref=25.0
):
    """
    Reweight a spectrum or spectrogram object.
    Input spectrogram assumes a shape of: N_frequencies x N_times
    This is meant to be a helper function used to change the spectral index of the stochastic results.

    Parameters
    ----------
        spec: array-like
            Spectrum or spectrogram (with shape N_frequencies x Ntimes)
        freqs: array-like
            Frequencies associated with `spec`.
        new_alpha: float
            New spectral index
        new_fref: float
            New reference frequency
        old_alpha: float, optional
            Spectral index of input `spec` array (i.e. weighting of `spec`). Defaults to zero (assumes unweighted)
        old_fref:
            Reference frequency of current `spec` weighting (assumes 1 Hz)

    Returns
    -------
        new_spec: array-like
            Reweighted spectrum or spectrogram array.
    """
    weights_old = (freqs / old_fref) ** old_alpha
    weights_new = (freqs / new_fref) ** new_alpha
    return (spec.T * (weights_old / weights_new)).T


register_hdf5_array_io(OmegaSpectrogram)
register_hdf5_array_io(OmegaSpectrum)
